/* ======================================================================== */
/*                                                                          */
/*  TEXAS INSTRUMENTS, INC.                                                 */
/*                                                                          */
/*  NAME                                                                    */
/*      forney_d -- Driver for forney functions, Little Endian              */
/*                                                                          */
/*  AUTHOR                                                                  */
/*      J. Random Coder                                                     */
/*                                                                          */
/*  REVISION HISTORY                                                        */
/*      12-Nov-1999 Initial Revision . . . . . . . . . . .  J.R. Coder      */
/*                                                                          */
/*  USAGE                                                                   */
/*      This code contains a driver program for testing the benchmark       */
/*      kernel.  The driver itself is invoked as a standard C program,      */
/*      and results are reported on stdout.                                 */
/*                                                                          */
/*      Building this with the "COMM_FLOW" and "RAND_DATA" macros set       */
/*      will build a driver which generates and tests with random data      */
/*      sets.  Otherwise, the driver will use a fixed test vector.          */
/*                                                                          */
/*  NOTES                                                                   */
/*      The various implementations of 'forney' are in the following        */
/*      files, with the following function names.                           */
/*                                                                          */
/*      -- Natural C                        forney_c.c - forney_cn()        */
/*      -- Optimized C                      forney_o.c - forney_co()        */
/*      -- Intrinsic C                      forney_i.c - forney_c()         */
/*      -- Non-partitioned Serial Assembly  forney_n.c - forney_nsa()       */
/*      -- Partitioned Serial Assembly      forney_p.c - forney_sa()        */
/*      -- Hand Optimized Assembly          forney_h.c - forney_asm()       */
/*                                                                          */
/* ------------------------------------------------------------------------ */
/*            Copyright (c) 1999 Texas Instruments, Incorporated.           */
/*                           All Rights Reserved.                           */
/* ======================================================================== */

static const char Copyright[] = "Copyright (C) 1999 Texas Instruments "
                                "Incorporated.  All Rights Reserved.";

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <limits.h>

#include "forney_c.h"
#include "forney_o.h"
#include "forney_i.h"
#include "forney_n.h"
#include "forney_p.h"
#include "forney_h.h"

#ifdef COMM_FLOW
# include "support.h"
#endif

#define SIZET     (8)
#define SIZE2T    (2*(SIZET))
#define RS_2T     SIZE2T
#define RS_T      SIZET
#define RS_N      (204)
#define DEGREE    (8)
#define PAD       (16)

const unsigned char log_table[256] =
{
    0xFF, 0x00, 0x01, 0x19, 0x02, 0x32, 0x1A, 0xC6,
    0x03, 0xDF, 0x33, 0xEE, 0x1B, 0x68, 0xC7, 0x4B,
    0x04, 0x64, 0xE0, 0x0E, 0x34, 0x8D, 0xEF, 0x81,
    0x1C, 0xC1, 0x69, 0xF8, 0xC8, 0x08, 0x4C, 0x71,
    0x05, 0x8A, 0x65, 0x2F, 0xE1, 0x24, 0x0F, 0x21,
    0x35, 0x93, 0x8E, 0xDA, 0xF0, 0x12, 0x82, 0x45,
    0x1D, 0xB5, 0xC2, 0x7D, 0x6A, 0x27, 0xF9, 0xB9,
    0xC9, 0x9A, 0x09, 0x78, 0x4D, 0xE4, 0x72, 0xA6,
    0x06, 0xBF, 0x8B, 0x62, 0x66, 0xDD, 0x30, 0xFD,
    0xE2, 0x98, 0x25, 0xB3, 0x10, 0x91, 0x22, 0x88,
    0x36, 0xD0, 0x94, 0xCE, 0x8F, 0x96, 0xDB, 0xBD,
    0xF1, 0xD2, 0x13, 0x5C, 0x83, 0x38, 0x46, 0x40,
    0x1E, 0x42, 0xB6, 0xA3, 0xC3, 0x48, 0x7E, 0x6E,
    0x6B, 0x3A, 0x28, 0x54, 0xFA, 0x85, 0xBA, 0x3D,
    0xCA, 0x5E, 0x9B, 0x9F, 0x0A, 0x15, 0x79, 0x2B,
    0x4E, 0xD4, 0xE5, 0xAC, 0x73, 0xF3, 0xA7, 0x57,
    0x07, 0x70, 0xC0, 0xF7, 0x8C, 0x80, 0x63, 0x0D,
    0x67, 0x4A, 0xDE, 0xED, 0x31, 0xC5, 0xFE, 0x18,
    0xE3, 0xA5, 0x99, 0x77, 0x26, 0xB8, 0xB4, 0x7C,
    0x11, 0x44, 0x92, 0xD9, 0x23, 0x20, 0x89, 0x2E,
    0x37, 0x3F, 0xD1, 0x5B, 0x95, 0xBC, 0xCF, 0xCD,
    0x90, 0x87, 0x97, 0xB2, 0xDC, 0xFC, 0xBE, 0x61,
    0xF2, 0x56, 0xD3, 0xAB, 0x14, 0x2A, 0x5D, 0x9E,
    0x84, 0x3C, 0x39, 0x53, 0x47, 0x6D, 0x41, 0xA2,
    0x1F, 0x2D, 0x43, 0xD8, 0xB7, 0x7B, 0xA4, 0x76,
    0xC4, 0x17, 0x49, 0xEC, 0x7F, 0x0C, 0x6F, 0xF6,
    0x6C, 0xA1, 0x3B, 0x52, 0x29, 0x9D, 0x55, 0xAA,
    0xFB, 0x60, 0x86, 0xB1, 0xBB, 0xCC, 0x3E, 0x5A,
    0xCB, 0x59, 0x5F, 0xB0, 0x9C, 0xA9, 0xA0, 0x51,
    0x0B, 0xF5, 0x16, 0xEB, 0x7A, 0x75, 0x2C, 0xD7,
    0x4F, 0xAE, 0xD5, 0xE9, 0xE6, 0xE7, 0xAD, 0xE8,
    0x74, 0xD6, 0xF4, 0xEA, 0xA8, 0x50, 0x58, 0xAF
};

const unsigned char exp_table[256] =
{
    0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,
    0x1D, 0x3A, 0x74, 0xE8, 0xCD, 0x87, 0x13, 0x26,
    0x4C, 0x98, 0x2D, 0x5A, 0xB4, 0x75, 0xEA, 0xC9,
    0x8F, 0x03, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0,
    0x9D, 0x27, 0x4E, 0x9C, 0x25, 0x4A, 0x94, 0x35,
    0x6A, 0xD4, 0xB5, 0x77, 0xEE, 0xC1, 0x9F, 0x23,
    0x46, 0x8C, 0x05, 0x0A, 0x14, 0x28, 0x50, 0xA0,
    0x5D, 0xBA, 0x69, 0xD2, 0xB9, 0x6F, 0xDE, 0xA1,
    0x5F, 0xBE, 0x61, 0xC2, 0x99, 0x2F, 0x5E, 0xBC,
    0x65, 0xCA, 0x89, 0x0F, 0x1E, 0x3C, 0x78, 0xF0,
    0xFD, 0xE7, 0xD3, 0xBB, 0x6B, 0xD6, 0xB1, 0x7F,
    0xFE, 0xE1, 0xDF, 0xA3, 0x5B, 0xB6, 0x71, 0xE2,
    0xD9, 0xAF, 0x43, 0x86, 0x11, 0x22, 0x44, 0x88,
    0x0D, 0x1A, 0x34, 0x68, 0xD0, 0xBD, 0x67, 0xCE,
    0x81, 0x1F, 0x3E, 0x7C, 0xF8, 0xED, 0xC7, 0x93,
    0x3B, 0x76, 0xEC, 0xC5, 0x97, 0x33, 0x66, 0xCC,
    0x85, 0x17, 0x2E, 0x5C, 0xB8, 0x6D, 0xDA, 0xA9,
    0x4F, 0x9E, 0x21, 0x42, 0x84, 0x15, 0x2A, 0x54,
    0xA8, 0x4D, 0x9A, 0x29, 0x52, 0xA4, 0x55, 0xAA,
    0x49, 0x92, 0x39, 0x72, 0xE4, 0xD5, 0xB7, 0x73,
    0xE6, 0xD1, 0xBF, 0x63, 0xC6, 0x91, 0x3F, 0x7E,
    0xFC, 0xE5, 0xD7, 0xB3, 0x7B, 0xF6, 0xF1, 0xFF,
    0xE3, 0xDB, 0xAB, 0x4B, 0x96, 0x31, 0x62, 0xC4,
    0x95, 0x37, 0x6E, 0xDC, 0xA5, 0x57, 0xAE, 0x41,
    0x82, 0x19, 0x32, 0x64, 0xC8, 0x8D, 0x07, 0x0E,
    0x1C, 0x38, 0x70, 0xE0, 0xDD, 0xA7, 0x53, 0xA6,
    0x51, 0xA2, 0x59, 0xB2, 0x79, 0xF2, 0xF9, 0xEF,
    0xC3, 0x9B, 0x2B, 0x56, 0xAC, 0x45, 0x8A, 0x09,
    0x12, 0x24, 0x48, 0x90, 0x3D, 0x7A, 0xF4, 0xF5,
    0xF7, 0xF3, 0xFB, 0xEB, 0xCB, 0x8B, 0x0B, 0x16,
    0x2C, 0x58, 0xB0, 0x7D, 0xFA, 0xE9, 0xCF, 0x83,
    0x1B, 0x36, 0x6C, 0xD8, 0xAD, 0x47, 0x8E, 0x00
};

const unsigned char exp_table2[512] =
{
    0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,
    0x1D, 0x3A, 0x74, 0xE8, 0xCD, 0x87, 0x13, 0x26,
    0x4C, 0x98, 0x2D, 0x5A, 0xB4, 0x75, 0xEA, 0xC9,
    0x8F, 0x03, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0,
    0x9D, 0x27, 0x4E, 0x9C, 0x25, 0x4A, 0x94, 0x35,
    0x6A, 0xD4, 0xB5, 0x77, 0xEE, 0xC1, 0x9F, 0x23,
    0x46, 0x8C, 0x05, 0x0A, 0x14, 0x28, 0x50, 0xA0,
    0x5D, 0xBA, 0x69, 0xD2, 0xB9, 0x6F, 0xDE, 0xA1,
    0x5F, 0xBE, 0x61, 0xC2, 0x99, 0x2F, 0x5E, 0xBC,
    0x65, 0xCA, 0x89, 0x0F, 0x1E, 0x3C, 0x78, 0xF0,
    0xFD, 0xE7, 0xD3, 0xBB, 0x6B, 0xD6, 0xB1, 0x7F,
    0xFE, 0xE1, 0xDF, 0xA3, 0x5B, 0xB6, 0x71, 0xE2,
    0xD9, 0xAF, 0x43, 0x86, 0x11, 0x22, 0x44, 0x88,
    0x0D, 0x1A, 0x34, 0x68, 0xD0, 0xBD, 0x67, 0xCE,
    0x81, 0x1F, 0x3E, 0x7C, 0xF8, 0xED, 0xC7, 0x93,
    0x3B, 0x76, 0xEC, 0xC5, 0x97, 0x33, 0x66, 0xCC,
    0x85, 0x17, 0x2E, 0x5C, 0xB8, 0x6D, 0xDA, 0xA9,
    0x4F, 0x9E, 0x21, 0x42, 0x84, 0x15, 0x2A, 0x54,
    0xA8, 0x4D, 0x9A, 0x29, 0x52, 0xA4, 0x55, 0xAA,
    0x49, 0x92, 0x39, 0x72, 0xE4, 0xD5, 0xB7, 0x73,
    0xE6, 0xD1, 0xBF, 0x63, 0xC6, 0x91, 0x3F, 0x7E,
    0xFC, 0xE5, 0xD7, 0xB3, 0x7B, 0xF6, 0xF1, 0xFF,
    0xE3, 0xDB, 0xAB, 0x4B, 0x96, 0x31, 0x62, 0xC4,
    0x95, 0x37, 0x6E, 0xDC, 0xA5, 0x57, 0xAE, 0x41,
    0x82, 0x19, 0x32, 0x64, 0xC8, 0x8D, 0x07, 0x0E,
    0x1C, 0x38, 0x70, 0xE0, 0xDD, 0xA7, 0x53, 0xA6,
    0x51, 0xA2, 0x59, 0xB2, 0x79, 0xF2, 0xF9, 0xEF,
    0xC3, 0x9B, 0x2B, 0x56, 0xAC, 0x45, 0x8A, 0x09,
    0x12, 0x24, 0x48, 0x90, 0x3D, 0x7A, 0xF4, 0xF5,
    0xF7, 0xF3, 0xFB, 0xEB, 0xCB, 0x8B, 0x0B, 0x16,
    0x2C, 0x58, 0xB0, 0x7D, 0xFA, 0xE9, 0xCF, 0x83,
    0x1B, 0x36, 0x6C, 0xD8, 0xAD, 0x47, 0x8E, 0x00,
    0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80,
    0x1D, 0x3A, 0x74, 0xE8, 0xCD, 0x87, 0x13, 0x26,
    0x4C, 0x98, 0x2D, 0x5A, 0xB4, 0x75, 0xEA, 0xC9,
    0x8F, 0x03, 0x06, 0x0C, 0x18, 0x30, 0x60, 0xC0,
    0x9D, 0x27, 0x4E, 0x9C, 0x25, 0x4A, 0x94, 0x35,
    0x6A, 0xD4, 0xB5, 0x77, 0xEE, 0xC1, 0x9F, 0x23,
    0x46, 0x8C, 0x05, 0x0A, 0x14, 0x28, 0x50, 0xA0,
    0x5D, 0xBA, 0x69, 0xD2, 0xB9, 0x6F, 0xDE, 0xA1,
    0x5F, 0xBE, 0x61, 0xC2, 0x99, 0x2F, 0x5E, 0xBC,
    0x65, 0xCA, 0x89, 0x0F, 0x1E, 0x3C, 0x78, 0xF0,
    0xFD, 0xE7, 0xD3, 0xBB, 0x6B, 0xD6, 0xB1, 0x7F,
    0xFE, 0xE1, 0xDF, 0xA3, 0x5B, 0xB6, 0x71, 0xE2,
    0xD9, 0xAF, 0x43, 0x86, 0x11, 0x22, 0x44, 0x88,
    0x0D, 0x1A, 0x34, 0x68, 0xD0, 0xBD, 0x67, 0xCE,
    0x81, 0x1F, 0x3E, 0x7C, 0xF8, 0xED, 0xC7, 0x93,
    0x3B, 0x76, 0xEC, 0xC5, 0x97, 0x33, 0x66, 0xCC,
    0x85, 0x17, 0x2E, 0x5C, 0xB8, 0x6D, 0xDA, 0xA9,
    0x4F, 0x9E, 0x21, 0x42, 0x84, 0x15, 0x2A, 0x54,
    0xA8, 0x4D, 0x9A, 0x29, 0x52, 0xA4, 0x55, 0xAA,
    0x49, 0x92, 0x39, 0x72, 0xE4, 0xD5, 0xB7, 0x73,
    0xE6, 0xD1, 0xBF, 0x63, 0xC6, 0x91, 0x3F, 0x7E,
    0xFC, 0xE5, 0xD7, 0xB3, 0x7B, 0xF6, 0xF1, 0xFF,
    0xE3, 0xDB, 0xAB, 0x4B, 0x96, 0x31, 0x62, 0xC4,
    0x95, 0x37, 0x6E, 0xDC, 0xA5, 0x57, 0xAE, 0x41,
    0x82, 0x19, 0x32, 0x64, 0xC8, 0x8D, 0x07, 0x0E,
    0x1C, 0x38, 0x70, 0xE0, 0xDD, 0xA7, 0x53, 0xA6,
    0x51, 0xA2, 0x59, 0xB2, 0x79, 0xF2, 0xF9, 0xEF,
    0xC3, 0x9B, 0x2B, 0x56, 0xAC, 0x45, 0x8A, 0x09,
    0x12, 0x24, 0x48, 0x90, 0x3D, 0x7A, 0xF4, 0xF5,
    0xF7, 0xF3, 0xFB, 0xEB, 0xCB, 0x8B, 0x0B, 0x16,
    0x2C, 0x58, 0xB0, 0x7D, 0xFA, 0xE9, 0xCF, 0x83,
    0x1B, 0x36, 0x6C, 0xD8, 0xAD, 0x47, 0x8E, 0x01
};

const unsigned char div_inv_table[256] =
{
    0x00, 0x01, 0x8E, 0xF4, 0x47, 0xA7, 0x7A, 0xBA,
    0xAD, 0x9D, 0xDD, 0x98, 0x3D, 0xAA, 0x5D, 0x96,
    0xD8, 0x72, 0xC0, 0x58, 0xE0, 0x3E, 0x4C, 0x66,
    0x90, 0xDE, 0x55, 0x80, 0xA0, 0x83, 0x4B, 0x2A,
    0x6C, 0xED, 0x39, 0x51, 0x60, 0x56, 0x2C, 0x8A,
    0x70, 0xD0, 0x1F, 0x4A, 0x26, 0x8B, 0x33, 0x6E,
    0x48, 0x89, 0x6F, 0x2E, 0xA4, 0xC3, 0x40, 0x5E,
    0x50, 0x22, 0xCF, 0xA9, 0xAB, 0x0C, 0x15, 0xE1,
    0x36, 0x5F, 0xF8, 0xD5, 0x92, 0x4E, 0xA6, 0x04,
    0x30, 0x88, 0x2B, 0x1E, 0x16, 0x67, 0x45, 0x93,
    0x38, 0x23, 0x68, 0x8C, 0x81, 0x1A, 0x25, 0x61,
    0x13, 0xC1, 0xCB, 0x63, 0x97, 0x0E, 0x37, 0x41,
    0x24, 0x57, 0xCA, 0x5B, 0xB9, 0xC4, 0x17, 0x4D,
    0x52, 0x8D, 0xEF, 0xB3, 0x20, 0xEC, 0x2F, 0x32,
    0x28, 0xD1, 0x11, 0xD9, 0xE9, 0xFB, 0xDA, 0x79,
    0xDB, 0x77, 0x06, 0xBB, 0x84, 0xCD, 0xFE, 0xFC,
    0x1B, 0x54, 0xA1, 0x1D, 0x7C, 0xCC, 0xE4, 0xB0,
    0x49, 0x31, 0x27, 0x2D, 0x53, 0x69, 0x02, 0xF5,
    0x18, 0xDF, 0x44, 0x4F, 0x9B, 0xBC, 0x0F, 0x5C,
    0x0B, 0xDC, 0xBD, 0x94, 0xAC, 0x09, 0xC7, 0xA2,
    0x1C, 0x82, 0x9F, 0xC6, 0x34, 0xC2, 0x46, 0x05,
    0xCE, 0x3B, 0x0D, 0x3C, 0x9C, 0x08, 0xBE, 0xB7,
    0x87, 0xE5, 0xEE, 0x6B, 0xEB, 0xF2, 0xBF, 0xAF,
    0xC5, 0x64, 0x07, 0x7B, 0x95, 0x9A, 0xAE, 0xB6,
    0x12, 0x59, 0xA5, 0x35, 0x65, 0xB8, 0xA3, 0x9E,
    0xD2, 0xF7, 0x62, 0x5A, 0x85, 0x7D, 0xA8, 0x3A,
    0x29, 0x71, 0xC8, 0xF6, 0xF9, 0x43, 0xD7, 0xD6,
    0x10, 0x73, 0x76, 0x78, 0x99, 0x0A, 0x19, 0x91,
    0x14, 0x3F, 0xE6, 0xF0, 0x86, 0xB1, 0xE2, 0xF1,
    0xFA, 0x74, 0xF3, 0xB4, 0x6D, 0x21, 0xB2, 0x6A,
    0xE3, 0xE7, 0xB5, 0xEA, 0x03, 0x8F, 0xD3, 0xC9,
    0x42, 0xD4, 0xE8, 0x75, 0x7F, 0xFF, 0x7E, 0xFD
};


/* ======================================================================== */
/*  Kernel-specific array alignment requirements.                           */
/* ======================================================================== */

#pragma DATA_ALIGN(lambda,      8);       /* Dword-aligned */
#pragma DATA_ALIGN(zeros,       8);       /* Dword-aligned */
#pragma DATA_ALIGN(s,           8);       /* Dword-aligned */
#pragma DATA_ALIGN(byte_i,      8);       /* Dword-aligned */
#pragma DATA_ALIGN(scratch_asm, 8);       /* Dword-aligned */
#pragma DATA_ALIGN(scratch_sa , 8);       /* Dword-aligned */
#pragma DATA_ALIGN(scratch_nsa, 8);       /* Dword-aligned */
#pragma DATA_ALIGN(scratch_cn , 8);       /* Dword-aligned */
#pragma DATA_ALIGN(scratch_co , 8);       /* Dword-aligned */
#pragma DATA_ALIGN(scratch_c  , 8);       /* Dword-aligned */
#pragma DATA_ALIGN(byte_i_asm,  8);       /* Dword-aligned */
#pragma DATA_ALIGN(byte_i_sa ,  8);       /* Dword-aligned */
#pragma DATA_ALIGN(byte_i_nsa,  8);       /* Dword-aligned */
#pragma DATA_ALIGN(byte_i_cn ,  8);       /* Dword-aligned */
#pragma DATA_ALIGN(byte_i_co ,  8);       /* Dword-aligned */
#pragma DATA_ALIGN(byte_i_c  ,  8);       /* Dword-aligned */

/* ======================================================================== */
/*  Global variables for test vectors.  If both COMM_FLOW and RAND_DATA     */
/*  are defined, use random test vectors, otherwise use fixed test vectors. */
/* ======================================================================== */

#if defined(COMM_FLOW) && defined(RAND_DATA)

/* ======================================================================== */
/*  Parameters for random vectors.                                          */
/* ======================================================================== */

/* ======================================================================== */
/*  Uninitialized arrays for random data.                                   */
/* ======================================================================== */

  unsigned char lambda[RS_T+1];
  unsigned char zeros[RS_T];
  unsigned char s[RS_2T];
  unsigned char byte_i[RS_N + 2*PAD];
  const unsigned char *tables[3] = {exp_table2, div_inv_table, log_table};

/* ======================================================================== */
/*  These variables are needed only when working with random test vectors.  */
/* ======================================================================== */
char *seed_fname = "seedfile";
int   seed;

#else


/* ======================================================================== */
/*  Initialized arrays with fixed test data.                                */
/* ======================================================================== */

const unsigned char lambda[] = 
{ 
    0x01, 0x02, 0xD7, 0x3C, 0x9C, 0x6A, 0x79, 0x71, 0x11 
};

const unsigned char zeros[]  = 
{ 
    0x5D, 0x81, 0xCC, 0x3F, 0xCC, 0x5D, 0xD1, 0xAF 
}; 

const unsigned char s[]      = 
{ 
    0x1F, 0xDF, 0x1E, 0x07, 0x90, 0x94, 0x45, 0xE3,
    0xDD, 0xB4, 0xFF, 0x22, 0x6E, 0x2C, 0xE3, 0x9C 
}; 

const unsigned char byte_i[RS_N + 2*PAD] =
{
    0xE2, 0x7A, 0x89, 0x02, 0x5A, 0xE6, 0x69, 0x26,   /* PAD */
    0xDF, 0xF9, 0x85, 0xEC, 0xC6, 0x0F, 0x65, 0x2E,   /* PAD */

    0xF4, 0xDC, 0xC9, 0x20, 0xB4, 0xC2, 0x87, 0x68, 
    0xCB, 0xE8, 0xFA, 0x08, 0x2D, 0xB7, 0xC2, 0x10, 
    0xC9, 0x4B, 0xB1, 0x57, 0xFB, 0x02, 0xE7, 0x39, 
    0x23, 0x6C, 0x07, 0x91, 0x16, 0x2E, 0x57, 0x3A, 
    0x07, 0x19, 0x80, 0x25, 0x4A, 0x81, 0x94, 0x1F, 
    0x42, 0xCC, 0xCD, 0x34, 0xE3, 0x23, 0x01, 0x33, 
    0x21, 0xD6, 0x48, 0xE9, 0x54, 0xBD, 0xB8, 0x28, 
    0x41, 0xE4, 0x1F, 0x9E, 0x91, 0xCB, 0xAF, 0xB1, 

    0xB2, 0x01, 0x81, 0x3F, 0xEB, 0x3D, 0xC5, 0x12, 
    0xB8, 0xD3, 0x1B, 0xB0, 0x09, 0x9D, 0x7D, 0xBF, 
    0xC5, 0x43, 0x8F, 0x4D, 0x5F, 0xC4, 0x3D, 0x9F, 
    0x7B, 0x00, 0x6A, 0x7D, 0xA6, 0x2D, 0x84, 0xFA, 
    0x3A, 0xF3, 0x54, 0x68, 0x34, 0x3D, 0xC1, 0x4B, 
    0x75, 0xCA, 0x91, 0xA1, 0xC7, 0xC4, 0x3C, 0x8B, 
    0xAC, 0xB1, 0x04, 0xC4, 0x79, 0xA3, 0xC4, 0xBA, 
    0x8E, 0x5D, 0xE8, 0x49, 0x69, 0x85, 0x8F, 0x02, 

    0xCE, 0x9F, 0xB2, 0x0C, 0x69, 0xB4, 0x38, 0x78, 
    0x27, 0xC6, 0xB5, 0x4C, 0x09, 0x30, 0x1B, 0x24, 
    0x2A, 0xD0, 0x6A, 0xC0, 0xAC, 0xC7, 0x6D, 0xDB, 
    0xB3, 0x2A, 0xA4, 0x76, 0x13, 0x3F, 0xAE, 0x78, 
    0x76, 0x6A, 0xE9, 0xA4, 0x67, 0xAA, 0x98, 0xEE, 
    0xDF, 0xF7, 0x98, 0x79, 0xE0, 0x83, 0x80, 0x5C, 
    0xE6, 0xF6, 0xF3, 0xF1, 0x45, 0x39, 0x06, 0xD1, 
    0xE6, 0x2B, 0x54, 0x6B, 0xDC, 0x4D, 0x4D, 0x42, 

    0x54, 0x83, 0xF8, 0xFB, 0x70, 0x09, 0x94, 0xD5, 
    0x16, 0x3F, 0xFF, 0x76, 

    0xFB, 0x77, 0xFB, 0x16, 0x99, 0xDA, 0x6C, 0x33,   /* PAD */
    0xA1, 0xA8, 0xD8, 0xC9, 0x33, 0x3B, 0xF7, 0x4E,   /* PAD */
};

const unsigned char *tables[3] = {exp_table2, div_inv_table, log_table};

#endif /* defined(COMM_FLOW) && defined(RAND_DATA) */

/* ======================================================================== */
/*  Result arrays, with padding to catch output overrun.                    */
/* ======================================================================== */

unsigned char byte_i_asm [RS_N + 2*PAD], *ptr_o_asm = byte_i_asm + PAD;
unsigned char byte_i_sa  [RS_N + 2*PAD], *ptr_o_sa  = byte_i_sa  + PAD;
unsigned char byte_i_nsa [RS_N + 2*PAD], *ptr_o_nsa = byte_i_nsa + PAD;
unsigned char byte_i_cn  [RS_N + 2*PAD], *ptr_o_cn  = byte_i_cn  + PAD;
unsigned char byte_i_co  [RS_N + 2*PAD], *ptr_o_co  = byte_i_co  + PAD;
unsigned char byte_i_c   [RS_N + 2*PAD], *ptr_o_c   = byte_i_c   + PAD;

unsigned char scratch_asm [11*RS_T + 2*PAD], *ptr_s_asm = scratch_asm + PAD;
unsigned char scratch_sa  [11*RS_T + 2*PAD], *ptr_s_sa  = scratch_sa  + PAD;
unsigned char scratch_nsa [11*RS_T + 2*PAD], *ptr_s_nsa = scratch_nsa + PAD;
unsigned char scratch_cn  [11*RS_T + 2*PAD], *ptr_s_cn  = scratch_cn  + PAD;
unsigned char scratch_co  [11*RS_T + 2*PAD], *ptr_s_co  = scratch_co  + PAD;
unsigned char scratch_c   [11*RS_T + 2*PAD], *ptr_s_c   = scratch_c   + PAD;

int fail_asm = 0;
int fail_sa  = 0;
int fail_nsa = 0;
int fail_c   = 0;
int fail_cn  = 0;
int fail_co  = 0;


/* ======================================================================== */
/*  Prototypes for timing functions.                                        */
/* ======================================================================== */
clock_t time_asm  ();
clock_t time_sa   ();
clock_t time_nsa  ();
clock_t time_c    ();
clock_t time_cn   ();
clock_t time_co   ();

/* ======================================================================== */
/*  MAIN -- Top level driver for the test.                                  */
/* ======================================================================== */
int main()
{
	FILE *f;
	char *fname = "stdout.txt";
    clock_t t_overhead, t_start, t_stop, t_asm, t_sa, t_nsa, t_c, t_cn, t_co;
    int fail = 0;

#if defined(COMM_FLOW) && defined(RAND_DATA)
    FILE *fp;

    /* -------------------------------------------------------------------- */
    /*  Load in a new random number seed from seedfile, or if seedfile      */
    /*  does not exist, pick an initial fixed seed.  Then seed rand().      */
    /* -------------------------------------------------------------------- */
    seed = 0x2A3A4A5A;
    if ((fp = fopen(seed_fname,"r")) != NULL)   /* open seed file   */
    {
        fscanf(fp,"%d",&seed);                  /* get current seed */
        fclose(fp);        
    }
    srand(seed);

    /* -------------------------------------------------------------------- */
    /*  Generate random test vector.                                        */
    /* -------------------------------------------------------------------- */

    randomize_char_array(byte_i, sizeof(byte_i), 1, 0, 255);
    randomize_char_array(s,      sizeof(s     ), 1, 0, 255);
    randomize_char_array(lambda, sizeof(lambda), 1, 0, 255);
    randomize_char_array(zeros,  sizeof(zeros ), 1, 0, 255);
    lambda[0] = 1;

#endif /* defined(COMM_FLOW) && defined(RAND_DATA) */

    /* -------------------------------------------------------------------- */
    /*  Fill output buffers with fixed values.                              */
    /* -------------------------------------------------------------------- */
    memset(scratch_asm, 0x00, sizeof(scratch_asm)/sizeof(unsigned char));
    memset(scratch_sa , 0x00, sizeof(scratch_asm)/sizeof(unsigned char));
    memset(scratch_nsa, 0x00, sizeof(scratch_asm)/sizeof(unsigned char));
    memset(scratch_cn , 0x00, sizeof(scratch_asm)/sizeof(unsigned char));
    memset(scratch_co , 0x00, sizeof(scratch_asm)/sizeof(unsigned char));
    memset(scratch_c  , 0x00, sizeof(scratch_asm)/sizeof(unsigned char));

    memcpy(byte_i_asm, byte_i, sizeof(byte_i)/sizeof(unsigned char));
    memcpy(byte_i_sa , byte_i, sizeof(byte_i)/sizeof(unsigned char));
    memcpy(byte_i_nsa, byte_i, sizeof(byte_i)/sizeof(unsigned char));
    memcpy(byte_i_cn , byte_i, sizeof(byte_i)/sizeof(unsigned char));
    memcpy(byte_i_co , byte_i, sizeof(byte_i)/sizeof(unsigned char));
    memcpy(byte_i_c  , byte_i, sizeof(byte_i)/sizeof(unsigned char));



    /* -------------------------------------------------------------------- */
    /*  Compute the overhead of calling clock() twice to get timing info.   */
    /* -------------------------------------------------------------------- */
    t_start    = clock();
    t_stop     = clock();
    t_overhead = t_stop - t_start + 6;

    /* -------------------------------------------------------------------- */
    /*  Call the individual timing routines, and subtract off overhead.     */
    /* -------------------------------------------------------------------- */


    t_asm  = time_asm()  - t_overhead;
    t_sa   = time_sa()   - t_overhead;
    t_nsa  = time_nsa()  - t_overhead;
    t_c    = time_c()    - t_overhead;
    t_cn   = time_cn()   - t_overhead;
    t_co   = time_co()   - t_overhead;

    /* -------------------------------------------------------------------- */
    /*  Print timing results.                                               */
    /* -------------------------------------------------------------------- */
	if (!(f = fopen(fname,"w")))  /* open output file */
  	{
  		printf("failure writing file \"%s\"\n", fname);
  		exit(1);
  	}

    fprintf(f,"K_forney: pSA: %d, nSA: %d, "
           "C: %d, natC: %d, optC: %d\n",
           t_sa, t_nsa, t_c, t_cn, t_co);

	fclose(f);
    /* -------------------------------------------------------------------- */
    /*  Check the results arrays, and report any failures.                  */
    /* -------------------------------------------------------------------- */

    if (t_asm && memcmp(byte_i_asm, byte_i_cn, sizeof(byte_i_cn)) != 0) 
        { printf("Result failure (asm)\n"); fail = 1; }

    if (t_sa  && memcmp(byte_i_sa,  byte_i_cn, sizeof(byte_i_cn)) != 0) 
        { printf("Result failure (sa)\n");  fail = 1; }

    if (t_nsa && memcmp(byte_i_nsa, byte_i_cn, sizeof(byte_i_cn)) != 0) 
        { printf("Result failure (nsa)\n"); fail = 1; }

    if (t_c   && memcmp(byte_i_c,   byte_i_cn, sizeof(byte_i_cn)) != 0) 
        { printf("Result failure (c)\n");   fail = 1; }

    if (t_co  && memcmp(byte_i_co,  byte_i_cn, sizeof(byte_i_cn)) != 0) 
        { printf("Result failure (co)\n");  fail = 1; }


    if (t_asm && memcmp(scratch_asm, scratch_cn, sizeof(scratch_cn)) != 0) 
        { printf("Result failure (asm)\n"); fail = 1; }

    if (t_sa  && memcmp(scratch_sa,  scratch_cn, sizeof(scratch_cn)) != 0) 
        { printf("Result failure (sa)\n");  fail = 1; }

    if (t_nsa && memcmp(scratch_nsa, scratch_cn, sizeof(scratch_cn)) != 0) 
        { printf("Result failure (nsa)\n"); fail = 1; }

    if (t_c   && memcmp(scratch_c,   scratch_cn, sizeof(scratch_cn)) != 0) 
        { printf("Result failure (c)\n");   fail = 1; }

    if (t_co  && memcmp(scratch_co,  scratch_cn, sizeof(scratch_cn)) != 0) 
        { printf("Result failure (co)\n");  fail = 1; }


    if (t_asm && memcmp(&fail_asm, &fail_cn, sizeof(&fail_cn)) != 0) 
        { printf("Result failure (asm)\n"); fail = 1; }

    if (t_sa  && memcmp(&fail_sa,  &fail_cn, sizeof(&fail_cn)) != 0) 
        { printf("Result failure (sa)\n");  fail = 1; }

    if (t_nsa && memcmp(&fail_nsa, &fail_cn, sizeof(&fail_cn)) != 0) 
        { printf("Result failure (nsa)\n"); fail = 1; }

    if (t_c   && memcmp(&fail_c,   &fail_cn, sizeof(&fail_cn)) != 0) 
        { printf("Result failure (c)\n");   fail = 1; }

    if (t_co  && memcmp(&fail_co,  &fail_cn, sizeof(&fail_cn)) != 0) 
        { printf("Result failure (co)\n");  fail = 1; }

#if defined(COMM_FLOW) && defined(RAND_DATA)
    /* -------------------------------------------------------------------- */
    /*  If the test succeeded, write a new seed to the seedfile, so that    */
    /*  the next random-data run will use an updated seed.  Otherwise,      */
    /*  leave the seed intact so that we can reproduce the error.           */
    /* -------------------------------------------------------------------- */
    if (!fail)
    {
        seed = rand();                      /* get next seed  */
        if (!(fp = fopen(seed_fname,"w")))  /* open seed file */
        {
            printf("failure writing seed file \"%s\"\n", seed_fname);
            exit(1);
        }
        fprintf(fp,"%d\n",seed);            /* set next seed  */
        fclose(fp);        
    }
#endif /* defined(COMM_FLOW) && defined(RAND_DATA) */

    return (fail);
}

/* ======================================================================== */
/*  TIME_ASM -- Measure elapsed time for hand-coded assembly version.       */
/* ======================================================================== */
clock_t time_asm()
{
    clock_t t_start, t_stop;

    t_start = clock();
#if !defined(COMM_FLOW) || !defined(NO_TIME_ASM)

    forney_asm( s, lambda, 8, zeros, ptr_o_asm,
               &fail_asm, tables, RS_T, RS_N, ptr_s_asm);

#endif
    t_stop  = clock();
    return  t_stop - t_start;
}

/* ======================================================================== */
/*  TIME_SA  -- Measure elapsed time for parititioned serial-assembly       */
/*              version.                                                    */
/* ======================================================================== */
clock_t time_sa()
{
    clock_t t_start, t_stop;

    t_start = clock();
#if !defined(COMM_FLOW) || !defined(NO_TIME_SA)

     forney_sa( s, lambda, 8, zeros, ptr_o_sa,
               &fail_sa, tables, RS_T, RS_N, ptr_s_sa);
#endif
    t_stop  = clock();
    return  t_stop - t_start;
}

/* ======================================================================== */
/*  TIME_NSA -- Measure elapsed time for non-parititioned serial-assembly   */
/*              version.                                                    */
/* ======================================================================== */
clock_t time_nsa()
{
    clock_t t_start, t_stop;

    t_start = clock();
#if !defined(COMM_FLOW) || !defined(NO_TIME_NSA)

     forney_nsa( s, lambda, 8, zeros, ptr_o_nsa,
               &fail_nsa, tables, RS_T, RS_N, ptr_s_nsa);
#endif
    t_stop  = clock();
    return  t_stop - t_start;
}

/* ======================================================================== */
/*  TIME_CO  -- Measure elapsed time for optimized (non-intrinsic) C        */
/*              version.                                                    */
/* ======================================================================== */
clock_t time_co()
{
    clock_t t_start, t_stop;

    t_start = clock();
#if !defined(COMM_FLOW) || !defined(NO_TIME_CO)

     forney_co( s, lambda, 8, zeros, ptr_o_co,
               &fail_co, tables, RS_T, RS_N, ptr_s_co);
#endif
    t_stop  = clock();
    return  t_stop - t_start;
}

/* ======================================================================== */
/*  TIME_C   -- Measure elapsed time for optimized C (with intrinsics)      */
/*              version.                                                    */
/* ======================================================================== */

clock_t time_c()
{
    clock_t t_start, t_stop;

    t_start = clock();
#if !defined(COMM_FLOW) || !defined(NO_TIME_C)

     forney_c( s, lambda, 8, zeros, ptr_o_c,
              &fail_c, tables, RS_T, RS_N, ptr_s_c);


#endif
    t_stop  = clock();
    return  t_stop - t_start;
}

/* ======================================================================== */
/*  TIME_CN  -- Measure elapsed time for natural C version.                 */
/* ======================================================================== */

clock_t time_cn()
{
    clock_t t_start, t_stop;

    t_start = clock();

    forney_cn( s, lambda, 8, zeros, ptr_o_cn,
             &fail_cn, tables, RS_T, RS_N, ptr_s_cn);

    t_stop  = clock();
    return  t_stop - t_start;
}

/* ======================================================================== */
/*  End of file:  forney_d.c                                              */
/* ------------------------------------------------------------------------ */
/*            Copyright (c) 1999 Texas Instruments, Incorporated.           */
/*                           All Rights Reserved.                           */
/* ======================================================================== */
