* ======================================================================== *
*  NAME                                                                    *
*                                                                          *
*      forney -- Forney Algorithm for Reed Solomon Decoder                 *
*                                                                          *
*S  AUTHOR                                                                 *
*S                                                                         *
*S      J. Sankaran                                                        *
*S                                                                         *
*S  REVISION HISTORY                                                       *
*S                                                                         *
*S      00-Jun-1998 Natural version of C code.............D.Hoyle          *
*S      10-Jan-2000 Initial revision of forney . . . . .  J.Sankaran       *
*S      03-10-2000  Fail condition 4 removed .............J.Sankaran       *
*                                                                          *
*  USAGE                                                                   *
*                                                                          *
*      This routine has the following C prototype:                         *
*                                                                          *
*    void forney_cn(  const  unsigned char restrict   s[],                 *
*                            unsigned char restrict   lambda[],            *
*                                     int             lam_deg,             *
*                            unsigned char restrict   zeros[],             *
*                            unsigned char restrict   byte_i[],            *
*                                     int             *fail,               *
*                      const unsigned char *restrict  tables[],            *
*                                     int             T,                   *
*                                     int             RS_N,                *
*                            unsigned char restrict   scratch[] )          *
*                                                                          *
*   T: # of errors that code can correct                                   *
*   s: array that contains 2T syndromes                                    *
*   lambda: array of error locator polynomial of degree T + 1              *
*           where lambda[0] = 1                                            *
*   lam_deg:Degree of the lambda polynomial.                               *
*   zeros:  Zeroes of the error locator polynomial found using Chien       *
*           search.                                                        *
*   byte_i: received code word of bytes                                    *
*   fail:   pointer to store status of correction                          *
*   tables: array of pointers that contain the foll: pointers              *
*           tables[0] ----> exp_table exponenentials of primitive element  *
*           tables[1] ----> log_table logarithms of the field elements     *
*   For eg: exp_table[8] = 2^8 = 29   and inversely log[29] = 8            *
*   for the case of (204,188,8) code for the default generator polynomial  *
*   100011101 GF(256)                                                      *
*   RS_N:   total number of bytes including parity bytes, 204 for the case *
*           of (204,188,8) code.                                           *
*   scratch: temporary scratch pad array of size 11 * T to fold temporary  *
*            results.                                                      *
*                                                                          *
*   The forney routine accepts the syndromes array "s" computed using the  *
*   syndrome routine, "lambda" error locator polynomial computed using the *
*   "Berlekamp" or equivalent, "zeros" array that contains the roots of    *
*   the error locator polynomial, "byte_i" the received code word and      *
*   corrects the errors in the received codeword array.                    *   
*                                                                          *
*  DESCRIPTION                                                             *
*                                                                          *
*  Let there be w errors, the zeroes of the lambda polynomial give these   *
*  locations. These are the roots at which the lambda polynomial evaluates *
*  to zero. However these are also the same locations where the error poly-*
*  nomial does not evaluate to zero. Therefore it can be written as foll:  *
*  ek . lambdak = 0 for k = 0,1,.....N-1 where N is the same as RS_N       *
*  Transforming to frequency domain Ek * Lambdak = 0 where * denotes       *
*  circular convolution. This implies that the dot product of every circul-*
*  arly shifted version of Ek with lambda polynomial evaluates to zero.    *
*  This circular convolution is evaluated over the known coefficients of E *
*  Since the calculations are performed on a field the linear convolution  *
*  ends up looking like a circular convolution.  Therefore the linear con- *
*  voultion equation may be written as: Ek*Lambdak mod(x^(N-1) -1) = 0     * 
*  This implies that the product is a multiple of x^n - 1, the multiple    *
*  being the error evaluator polynomial Omega. Therefore we may write:     *
*                                                                          *
*  Ek(x) * Lambdak(x) = Omegak(x).(x^N - 1)                                *
*                                                                          *
*  The forney algorithm calculates the first T values of Omega. Knowing the*
*  first T values of omega if the number of errors is <= T, coefficients of*
*  the error polynomial can be evaluated at these w locations yielding     *
*                                                                          *
*  ek = 1/N [(Omega(x)(x^N-1)/Lambda(x)]| for x= alpha^1, alpha^w          *
*                                                                          *
*  However these are the same locations at which Lambda(x) is zero, hence  *
*  the expresssion becomes indeterminate. Therefore by applying L'Hospital *
*  's rule yields:                                                         *
*                                                                          *
*  ek = [alpha^k * Omega(alpha^-k)]/ Lambda'(alpha^-k)                     *
*                                                                          *
*  where                                                                   *
*        ek: error magnitude                                               *
*        alpha: primitive element                                          *
*        Omega: Error evaluator polynomial                                 *
*        Lambda': Derivative of the Lambda polymomial                      *
*        alpha^-k: zeroes or locations of the errors of the error locator  *
*                  polynomial                                              *
*                                                                          *
*  Error positions are given by log[1/alpha^-k]. Knowing the error locns:  *
*  and the error magnitudes the errors can be corrected by XOR ing the     *
*  the error that was added at location k with the error magnitude ek      *
*                                                                          *
*                                                                          *
*  TECHNIQUES                                                              *
*                                                                          *
*  Galois field addition is done using XOR's as before, while the multi-   *
*  plication is done using the GMPY4 instruction. This helps to minimize   *
*  the difficulty in developing the code.                                  *
*  The omega polynomial calculation can be completely unrolled. This cal-  *
*  culation amounts to a circular convolution between the lambda poly-     *
*  nomial and the syndrome. This can be seen by writing out the omega      *
*  terms explicitly as follows:                                            *
*                                                                          *
*  Omega0 = lam0 * s0                                                      *
*  Omega1 = lam1 * s0  + lam0 * s1                                         *
*  Omega2 = lam2 * s0  + lam1 * s1 + lam0 * s2                             *
*  Omega3 = lam3 * s0  + lam2 * s1 + lam1 * s1 + lam0 * s3                 *
*  OmegaT = lamT * s0  +....                   + lam0 * sT                 *
*                                                                          *
*  Although in theory 2T syndromes are available, only T terms need to be  *
*  calculated, as the omega polynomial is the error  evaluator polynomial  *
*  and since there can be utmost T errors T terms of omega suffice.        *
*                                                                          *
*  The derivative of the Lambda polynomial is merely the shifted version   *
*  of the lambda polynomial zeroed out at the even locations. This is true *
*  because the finite field is of the form GF(2^m). Therefore even terms   *
*  like 2 evaluate to zero.                                                *
*                                                                          *
*  The numerator for the kth error is the omega polynomial evaluated at    *
*  kth zero of the lambda polynomial and is Omega(alpha^-k). The denomi-   *
*  nator on the other hand is the derivative of the lambda polynomial      *
*  evaluated at the kth zero. It is the derivative and not the lambda      *
*  polynomial itself as by definition the lambda polynomial would          *
*  evaluate to zero at the kth zero and division by zero is not defined,   *
*  even in finite fields. The multiplication by alpha^k with the           *
*  numerator is instead of multiplying the denominator by the zero.        *
*                                                                          *
*  ASSUMPTIONS                                                             *
*                                                                          *
* a) s: array of syndromes of size 2T                                      *
* b) lambda: error locator polynomial computed by Berlekamp or equiv.      *
* c) lam_deg: degree of error locator polynomial.                          *
* d) zeros: zeroes of the error locator polynomial found by Chien search.  *
* e) byte_i: received code word of size N                                  *
* f) fail: location where fail status is stored                            *
* g) None of these arrays overlp in memory.                                *
*                                                                          *
*                                                                          *
*  NOTES                                                                   *
*                                                                          *
*  None                                                                    *
*                                                                          *
*  MEMORY NOTE                                                             *
*                                                                          *
*  This benchmark is Little Endian. No bank conflicts should occur in this *
*  code.                                                                   *
*                                                                          *
*  CYCLES                                                                  *
*           T is the # errors that the code can correct.                   *
*           Loop1: 5[T-1] + 14 cycles  Loop2: 7[T-1] + 11 cycles           *
*           Additional: 37 cycles                                          *
*  T = 8    Loop1: 49 cycles Loop2: 60 cycles Additional: 37 cyles         *
*           Total = 146 cycles                                             *
*           Total = 12T + 50 cycles                                        *
*                                                                          * 
*S  SOURCE                                                                 *
*S                                                                         *
*S The natural form of C code borrows ideas and concepts derived by the    *
*S following individuals:                                                  *
*S                                                                         *
*S a) Tod Wolf: DSP R&D                                                    *
*S b) David Hoyle: DSP C6000 Software apps group.                          *
*                                                                          *
*  C code                                                                  *
*                                                                          *
* The following is the equivalent C code. Certain restrictions may apply:  *
*                                                                          *
*                                                                          *
*  void forney  (  const  unsigned char *restrict  s,                      *
*                          unsigned char *restrict  lambda,                *
*                                   int             lam_deg,               *
*                          unsigned char *restrict  zeros,                 *
*                          unsigned char *restrict  byte_i,                *
*                                   int             *fail,                 *
*                    const unsigned char *restrict  *restrict tables,      *
*                                   int             T,                     *
*                                   int             RS_N,                  *
*                          unsigned char *restrict  scratch )              *
*                                                                          *
*  {                                                                       *
*                                                                          *
*     int i,j,k;                                                           *
*                                                                          *
*                                                                          *
*    unsigned char *omega = scratch;                                       *
*    unsigned char *syn   = omega  +  3*T;                                 *
*    unsigned char *numerator = syn + 4*T;                                 *
*    unsigned char *denominator = numerator + T;                           *
*    unsigned char *poly  = denominator   +  T;                            *
*    unsigned char err_pos_i, err_mag_i, b;                                *
*    unsigned char constant = 0xFF;                                        *
*                                                                          *
*    unsigned int lam_t20, lam_word0;                                      *
*    unsigned int lam_t31, lam_word1;                                      *
*    unsigned int lam_t64, lam_word2;                                      *
*    unsigned int lam_t75, lam_word3;                                      *
*    unsigned int lam_word4, lam_word00;                                   *
*    unsigned int lam_word5, lam_word01;                                   *
*    unsigned int lam_word6;                                               *
*    unsigned int lam_word7;                                               *
*                                                                          *
*    unsigned int *syn_ptr;                                                *
*    unsigned int sword3210, sword7654;                                    *
*                                                                          *
*    unsigned int statew0, statew1;                                        *
*    unsigned int statew2, statew3;                                        *
*    unsigned int statew4, statew5;                                        *
*    unsigned int statew6, statew7;                                        *
*    unsigned int statew8, statew9;                                        *
*    unsigned int statew10, statew11;                                      *
*                                                                          *
*    unsigned int *omega_ptr = (unsigned int *) (omega);                   *
*    unsigned int omega_word0, omega_word1;                                *
*                                                                          *
*    unsigned int *lamptr = (unsigned int *) (lambda);                     *
*    unsigned int *poly_ptr = (unsigned int *) (poly);                     *
*    unsigned int  poly_word0, poly_word1, poly_word2, poly_word3;         *
*                                                                          *
*    int  RS_T = T;                                                        *
*                                                                          *
*    const unsigned char *div_inv_table = tables[1];                       *
*    const unsigned char *log_table = tables[2];                           *
*                                                                          *
*    double *zerosptr, zero_val;                                           *
*    unsigned int zeroword0, zeroword1;                                    *
*    unsigned int denword0,  denword1;                                     *
*    unsigned int numword0, numword1;                                      *
*                                                                          *
*    unsigned char *ptr_omega, *ptr_poly;                                  *
*    unsigned int   polyval, polyword, omegaval, omegaword;                *
*    unsigned int  denval, numval, zerval;                                 *
*    unsigned int  denom_i, numer_i, zero_i;                               *
*                                                                          *
*    lam_word00 = *lamptr++;                                               *
*    lam_word01 = *lamptr++;                                               *
*                                                                          *
*    lam_t20 = _pack2(lam_word00,  lam_word00);                            *
*    lam_t31 = _packh2(lam_word00, lam_word00);                            *
*    lam_t64 = _pack2(lam_word01,  lam_word01);                            *
*    lam_t75 = _packh2(lam_word01, lam_word01);                            *
*                                                                          *
*    lam_word0 = _packl4(lam_t20, lam_t20);                                *
*    lam_word1 = _packh4(lam_t20, lam_t20);                                *
*    lam_word2 = _packl4(lam_t31, lam_t31);                                *
*    lam_word3 = _packh4(lam_t31, lam_t31);                                *
*    lam_word4 = _packl4(lam_t64, lam_t64);                                *
*    lam_word5 = _packh4(lam_t64, lam_t64);                                *
*    lam_word6 = _packl4(lam_t75, lam_t75);                                *
*    lam_word7 = _packh4(lam_t75, lam_t75);                                *
*                                                                          *
*    syn_ptr = (unsigned int *) (s);                                       *
*    sword3210   = *syn_ptr++;                                             *
*    sword7654   = *syn_ptr++;                                             *
*                                                                          *
*    statew0     = _gmpy4(lam_word0, sword3210);                           *
*    statew1     = _gmpy4(lam_word0, sword7654);                           *
*                                                                          *
*    sword7654   = _shlmb(sword3210, sword7654);                           *
*    sword3210   <<= 8;                                                    *
*                                                                          *
*    statew2     = _gmpy4(lam_word1, sword3210);                           *
*    statew3     = _gmpy4(lam_word1, sword7654);                           *
*                                                                          *
*    sword7654   = _shlmb(sword3210, sword7654);                           *
*    sword3210   <<= 8;                                                    *
*                                                                          *
*    statew4     = _gmpy4(lam_word2, sword3210);                           *
*    statew5     = _gmpy4(lam_word2, sword7654);                           *
*                                                                          *
*    sword7654   = _shlmb(sword3210, sword7654);                           *
*    sword3210   <<= 8;                                                    *
*                                                                          *
*    statew6     = _gmpy4(lam_word3, sword3210);                           *
*    statew7     = _gmpy4(lam_word3, sword7654);                           *
*                                                                          *
*    sword7654   = _shlmb(sword3210, sword7654);                           *
*    sword3210   <<= 8;                                                    *
*                                                                          *
*    statew8     = _gmpy4(lam_word4, sword7654);                           *
*                                                                          *
*    sword7654   <<= 8;                                                    *
*                                                                          *
*    statew9     = _gmpy4(lam_word5, sword7654);                           *
*                                                                          *
*    sword7654   <<= 8;                                                    *
*                                                                          *
*    statew10    = _gmpy4(lam_word6, sword7654);                           *
*                                                                          *
*    sword7654   <<= 8;                                                    *
*                                                                          *
*    statew11     = _gmpy4(lam_word7, sword7654);                          *
*                                                                          *
*    omega_word0  = statew0 ^ statew2 ^ statew4 ^ statew6;                 *
*    omega_word1  = statew1 ^ statew3 ^ statew5 ^ statew7 ^                *
*                 statew8 ^ statew9 ^ statew10 ^ statew11;                 *
*                                                                          *
*    *omega_ptr++ = omega_word0;                                           *
*    *omega_ptr++ = omega_word1;                                           *
*                                                                          *
*                                                                          *
*    poly_word3 = 0;                                                       *
*    poly_word2 = 0;                                                       *
*                                                                          *
*    poly_word0 = (lam_word00 & 0xFF00FF00);                               *
*    poly_word1 = (lam_word01 & 0xFF00FF00);                               *
*                                                                          *
*                                                                          *
*    *poly_ptr++ = poly_word0;                                             *
*    *poly_ptr++ = poly_word1;                                             *
*    *poly_ptr++ = poly_word2;                                             *
*    *poly_ptr++ = poly_word3;                                             *
*                                                                          *
*                                                                          *
*    zerosptr = (double *) (zeros);                                        *
*    zero_val = *zerosptr;                                                 *
*    zeroword0 = _lo(zero_val);                                            *
*    zeroword1 = _hi(zero_val);                                            *
*                                                                          *
*    denword1 = denword0 = 0;                                              *
*    numword1 = numword0 = 0;                                              *
*                                                                          *
*    ptr_omega = omega + 7;                                                *
*    ptr_poly  = poly  + 8;                                                *
*                                                                          *
*    _nassert(RS_T == 8);                                                  *
*                                                                          *
*    for (j = RS_T-1; j >= 0; j--)                                         *
*    {                                                                     *
*         polyval=*ptr_poly --;                                            *
*         polyword=_pack2(polyval,polyval);                                *
*         polyword=_packl4(polyword,polyword);                             *
*                                                                          *
*         omegaval=*ptr_omega --;                                          *
*         omegaword=_pack2(omegaval,omegaval);                             *
*         omegaword=_packl4(omegaword,omegaword);                          *
*                                                                          *
*         denword0=_gmpy4(denword0,zeroword0);                             *
*         denword0=(denword0^polyword);                                    *
*         numword0=_gmpy4(numword0,zeroword0);                             *
*         numword0=(numword0^omegaword);                                   *
*         denword1=_gmpy4(denword1,zeroword1);                             *
*         denword1=(denword1^polyword);                                    *
*         numword1=_gmpy4(numword1,zeroword1);                             *
*         numword1=(numword1^omegaword);                                   *
*    }                                                                     *
*                                                                          *
*    denword0 = _gmpy4(denword0,zeroword0);                                *
*    denword1 = _gmpy4(denword1,zeroword1);                                *
*                                                                          *
*    j = 4;                                                                *
*    denom_i = denword0;                                                   *
*    zero_i  = zeroword0;                                                  *
*    numer_i = numword0;                                                   *
*                                                                          *
*    for(i = 0;i < T; i++)                                                 *
*    {                                                                     *
*        denval = (denom_i & constant);                                    *
*        zerval = (zero_i  & constant);                                    *
*        numval = (numer_i & constant);                                    *
*                                                                          *
*        denom_i = (denom_i >> 8);                                         *
*        zero_i  = (zero_i >> 8);                                          *
*        numer_i = (numer_i >> 8);                                         *
*                                                                          *
*        b = div_inv_table[denval];                                        *
*        j--;                                                              *
*        if(!j) denom_i = denword1;                                        *
*        if(!j) numer_i = numword1;                                        *
*        if(!j) zero_i  = zeroword1;                                       *
*        if(!j) j = 4;                                                     *
*                                                                          *
*        err_mag_i = _gmpy4( numval, b);                                   *
*        err_pos_i = log_table[div_inv_table[zerval]];                     *
*                                                                          *
*        if (!denval)  *fail = 4;                                          *
*        if (err_pos_i > RS_N-1)  *fail = 5;                               *
*                                                                          *
*        k = RS_N-1 - err_pos_i;                                           *
*        if (!*fail) byte_i[k] = ( byte_i[k]^err_mag_i );                  *
*   }                                                                      *
*                                                                          *
*                                                                          *
*}                                                                         *
*                                                                          *
*                                                                          *
*  CODESIZE                                                                *
*         The known codesize of this file is 792 bytes                     *
* ======================================================================== *
*             Copyright (c) 1999 Texas Instruments, Incorporated.          *
*                            All Rights Reserved.                          *
* ======================================================================== *

                .sect ".data:copyright_h"
_Copyright:     .string "Copyright (C) 1999 Texas Instruments Incorporated. "
                .string "All Rights Reserved.",0
                .sect ".text:hand"
                .include "forney_h.h62"

_forney_asm:

*S---------------------------------------------------------------------------*
*S                TI Proprietary Information:  Internal Data                 *
*S---------------------------------------------------------------------------*
*S Swipe6ng datecode: 000104 (04:19:14 PM) [REL]
*S Swipe6ng commandline:
*S swipe6ng -Ma -s -v -l -mv6400 prol.sa
*S Target architecture: "C64xx ISA"
*S---------------------------------------------------------------------------*
*S length = 31
*S Total register usage: 30 (a:13 b:17)
; ====================== SYMBOLIC REGISTER ASSIGNMENTS =======================
        .asg            A4,         A_s                         ; 
        .asg            B4,         B_lambda                    ; 
        .asg            A6,         A_lam_deg                   ; 
        .asg            B6,         B_zeros                     ; 
        .asg            A8,         A_byte_i                    ; 
        .asg            B8,         B_fail                      ; 
        .asg            A10,        A_tables                    ; 
        .asg            B10,        B_T                         ; 
        .asg            A12,        A_N                         ; 
        .asg            B12,        B_scratch                   ; 
        .asg            B27,        B_ptr_poly                  ; 
        .asg            B25,        B_ptr_omega                 ; 
        .asg            B17,        B_denword0                  ; 
        .asg            B16,        B_numword0                  ; 
        .asg            A9,         A_denword1                  ; 
        .asg            A7,         A_numword1                  ; 
        .asg            B20,        B_zeroword0                 ; 
        .asg            A3,         A_zeroword1                 ; 
        .asg            B0,         B_iters                     ; 
        .asg            B7,         B_lam_word00
        .asg            A7,         A_lam_word01
        .asg            B0,         B_lam_t20
        .asg            B1,         B_lam_t31
        .asg            A1,         A_lam_t64
        .asg            A0,         A_lam_t75
        .asg            B1,         B_lam_word0
        .asg            B5,         B_lam_word1
        .asg            B1,         B_lam_word2
        .asg            B0,         B_lam_word3
        .asg            A2,         A_lam_word4
        .asg            A1,         A_lam_word5
        .asg            A1,         A_lam_word6
        .asg            A0,         A_lam_word7
        .asg            B17,        B_s7654
        .asg            B16,        B_s3210
        .asg            B9,         B_statew0
        .asg            B2,         B_statew1
        .asg            B2,         B_statew2
        .asg            B20,        B_statew3
        .asg            A16,        A_s3210
        .asg            B21,        B_statew4
        .asg            B1,         B_statew5
        .asg            A3,         A_s7654
        .asg            A2,         A_lam_word3
        .asg            A2,         A_statew6
        .asg            A1,         A_statew7
        .asg            A0,         A_statew8
        .asg            A3,         A_statew9
        .asg            A5,         A_statew10
        .asg            A9,         A_statew11
        .asg            B2,         B_statew20
        .asg            B1,         B_statew64
        .asg            B16,        B_omgword0
        .asg            B7,         B_statew31
        .asg            A0,         A_statew75
        .asg            A1,         A_statew89
        .asg            B7,         B_statew35
        .asg            A1,         A_statew80
        .asg            B17,        B_omgword1
        .asg            B1,         B_omega
        .asg            B19,        B_polyword3
        .asg            B18,        B_polyword2
        .asg            B0,         B_constant
        .asg            A0,         A_constant
        .asg            B5,         B_poly_ptr
        .asg            B16,        B_polyword0
        .asg            B17,        B_polyword1
        .asg            B21,        B_zeroword1
        .asg            B9,         B_poly_val
        .asg            B30,        B_csr
        .asg            B31,        B_no_gie
; ============================================================================

        LDW     .D2T2   *B_lambda[0],   B_lam_word00                  ; lambda[0]
||      MVC     .S2     CSR,            B_csr                         ; CSR

        NOP             1                                             ;
||      AND             B_csr,      -2,         B_no_gie              ; Clear GIE

        LDDW    .D1T2   *A_s,           B_s7654:B_s3210               ; Syndromes
||      MVC     .S2     B_no_gie,       CSR                           ; Dis. ints

        NOP             2                                             ; 

        PACK2   .S2     B_lam_word00,   B_lam_word00,    B_lam_t20    ; Pack lam0

        PACKL4  .L2     B_lam_t20,      B_lam_t20,       B_lam_word0  ; lam_word0

        GMPY4   .M2     B_lam_word0,    B_s7654,         B_statew1    ; lam0*s7654 
||      SHLMB   .L2     B_s3210,        B_s7654,         B_s7654      ; Shift.poly

        LDW     .D2T1   *B_lambda[1],   A_lam_word01                  ; lam_word1 
||      GMPY4   .M2     B_lam_word0,    B_s3210,         B_statew0    ; statew0
||      SHL     .S2     B_s3210,        8,               B_s3210      ; Shift.poly
||      PACKH4  .L2     B_lam_t20,      B_lam_t20,       B_lam_word1  ; lam_word1

        PACKH2  .L2     B_lam_word00,   B_lam_word00,    B_lam_t31    ; lam_t31
||      GMPY4   .M2     B_lam_word1,    B_s7654,         B_statew3    ; statew3
||      SHLMB   .S2     B_s3210,        B_s7654,         B_s7654      ; Shift.poly

        SHL     .S1X    B_s3210,        8,               A_s3210      ; s3210
||      PACKH4  .L2     B_lam_t31,      B_lam_t31,       B_lam_word3  ; lam_word3

        PACKL4  .L2     B_lam_t31,      B_lam_t31,       B_lam_word2  ; lam_word2
||      SHLMB   .S1X    A_s3210,        B_s7654,         A_s7654      ; Shift.poly

        GMPY4   .M2X    B_lam_word2,    A_s3210,         B_statew4    ; statew4
||      SHL     .S1     A_s3210,        8,               A_s3210      ; s3210
||      MV      .L1X    B_lam_word3,    A_lam_word3                   ; lam_word3

        GMPY4   .M1     A_lam_word3,    A_s3210,         A_statew6    ; statew6
||      PACKH2  .S1     A_lam_word01,   A_lam_word01,    A_lam_t75    ; lam_t75
||      PACK2   .L1     A_lam_word01,   A_lam_word01,    A_lam_t64    ; lam_t64

        GMPY4   .M1     A_lam_word3,    A_s7654,         A_statew7    ; statew7
||      PACKL4  .L1     A_lam_t64,      A_lam_t64,       A_lam_word4  ; lam_word4
||      SHLMB   .S1     A_s3210,        A_s7654,         A_s7654      ; s7654

        PACKH4  .L1     A_lam_t64,      A_lam_t64,       A_lam_word5  ; lam_word5
||      GMPY4   .M1     A_lam_word4,    A_s7654,         A_statew8    ; statew8
||      SHL     .S1     A_s7654,        8,               A_s7654      ; s7654

        PACKL4  .L1     A_lam_t75,      A_lam_t75,       A_lam_word6  ; lam_word6
||      GMPY4   .M1     A_lam_word5,    A_s7654,         A_statew9    ; statew9
||      SHL     .S1     A_s7654,        8,               A_s7654      ; s7654

        GMPY4   .M2     B_lam_word2,    B_s7654,         B_statew5    ; statew5
||      GMPY4   .M1     A_lam_word6,    A_s7654,         A_statew10   ; statew10
||      SHL     .S1     A_s7654,        8,               A_s7654      ; s7654
||      PACKH4  .L1     A_lam_t75,      A_lam_t75,       A_lam_word7  ; lam_word7

        GMPY4   .M1     A_lam_word7,    A_s7654,         A_statew11   ; statew11
||      ADD     .L2     B_scratch,      7,               B_ptr_omega  ; omega ptr

        ZERO    .D2     B_polyword2                                   ; poly_word2

        ZERO    .L2     B_polyword3                                   ; poly_word3
||      ADD     .D2     B_scratch,       B_T,            B_poly_ptr   ; poly_ptr
||      MVKL    .S2     0FF00FF00h,      B_constant                   ; constant
||      GMPY4   .M2     B_lam_word1,     B_s3210,        B_statew2    ; statew2

        STDW    .D2T2   B_polyword3:B_polyword2,        *B_poly_ptr[1]; store poly
||      MVKH    .S2     0FF00FF00h,      B_constant                   ; constant
||      ADD     .L2     B_poly_ptr,      8,              B_ptr_poly   ; Twin ptr

        AND     .L2X    A_lam_word01,    B_constant,     B_polyword1  ; polyword1
||      XOR     .S1X    A_statew7,       B_statew5,      A_statew75   ; sw7^sw5
||      XOR     .L1     A_statew10,      A_statew11,     A_statew10   ; sw10^sw11
||      XOR     .D1     A_statew8,       A_statew9,      A_statew89   ; sw8^sw9
||      LDDW    .D2T2   *B_zeros,        B_zeroword1:B_zeroword0      ; zeroes

        AND     .D2     B_lam_word00,    B_constant,      B_polyword0 ; polyword0
||      XOR     .L2X    B_statew4,       A_statew6,       B_statew64  ; sw4^sw6
||      XOR     .S2     B_statew1,       B_statew3,       B_statew31  ; sw1^sw3
||      XOR     .D1     A_statew89,      A_statew10,      A_statew80  ; sw89^sw10

        STDW    .D2T2   B_polyword1:B_polyword0,        *B_poly_ptr[0]; Store poly
||      XOR     .S2     B_statew0,       B_statew2,      B_statew20   ; sw0^sw2
||      XOR     .L2X    B_statew31,      A_statew75,     B_statew35   ; sw31^sw75

        XOR     .S2     B_statew20,      B_statew64,      B_omgword0  ; Omegaword0
||      XOR     .L2X    A_statew80,      B_statew35,      B_omgword1  ; Omegaword1

        MV      .L1X    B_constant,      A_constant                   ; Constant
||      STDW    .D2T2   B_omgword1:B_omgword0,            *B_scratch  ; Store 
||      SHL     .S1     A_s3210,         8,               A_s3210     ; s3210
||      SUB     .S2     B_T,             1,               B_iters     ; T - 1
||      ZERO    .L2     B_numword0                                    ; numword0

        MV      .S1X    B_zeroword1,     A_zeroword1                  ; Copy
||      ZERO    .D1     A_numword1                                    ; numword1
||      ZERO    .L1     A_denword1                                    ; denword1
||      ZERO    .S2     B_denword0                                    ; denword0
||      LDBU    .D2T2   *B_ptr_poly--,   B_poly_val                   ; poly_val

        ;==== Branch occurs  

*S---------------------------------------------------------------------------*
*S                TI Proprietary Information:  Internal Data                 *
*S---------------------------------------------------------------------------*
*S Swipe6ng datecode: 000104 (04:19:14 PM) [REL]
*S Swipe6ng commandline:
*S swipe6ng -Ma -s -v -2 -mv6400 loop1.cmd
*S Target architecture: "C64xx ISA"
*S---------------------------------------------------------------------------*
*S
*S Resource MII: 3
*S
*S Resource: L1         L1-unit hardware                           Bound: 2
*S Resource: S1         S1-unit hardware                           Bound: 1
*S Resource: D1         D1-unit hardware                           Bound: 1
*S Resource: M1         M1-unit hardware                           Bound: 2
*S Resource: T1         T1-unit LD/ST transport (DAB1)             Bound: 1
*S Resource: L2         L2-unit hardware                           Bound: 2
*S Resource: S2         S2-unit hardware                         * Bound: 3
*S Resource: D2         D2-unit hardware                         * Bound: 3
*S Resource: M2         M2-unit hardware                           Bound: 2
*S Resource: T2         T2-unit LD/ST transport (DAB2)             Bound: 1
*S Resource: M1l        M1-unit write port (lo half)               Bound: 2
*S Resource: M2l        M2-unit write port (lo half)               Bound: 2
*S Resource: xoper1     Unshared cross path on 1X                  Bound: 1
*S Resource: xoper2     Unshared cross path on 2X                  Bound: 1
*S Resource: slots      Instruction issue slots: 16                Bound: 2
*S
*S Recurrence MII: 5
*S
*S---------------------------------------------------------------------------*
*S Running adjust phase w/ register allocator
*S---------------------------------------------------------------------------*
*S File   : loop1.cmd
*S ResMII : 3
*S RecMII : 5
*S MII    : 5
*S II     : 5
*S No ILP : 0
*S List   : 13
*S Length : 11
*S Start  : 0
*S Stop   : 10
*S Operations         : 16
*S Avg. para @ II     : 3.200000
*S Calc. Operations   : 16
*S Avg. para calc @ II: 3.200000
*S Live Too Long : 0
*S Unroll Factor : 0
*S---------------------------------------------------------------------------*
*S Total register usage: 11 (a:4 b:7)
; ====================== SYMBOLIC REGISTER ASSIGNMENTS =======================
        .asg            B27,        B_ptr_poly
        .asg            B25,        B_ptr_omega
        .asg            B17,        B_denword0
        .asg            B16,        B_numword0
        .asg            A9,         A_denword1
        .asg            A7,         A_numword1
        .asg            B20,        B_zeroword0                 ; 
        .asg            A3,         A_zeroword1
        .asg            B0,         B_iters
        .asg            B9,         B_poly_val
        .asg            B9,         B_poly_word
        .asg            A5,         A_omega_val
        .asg            A5,         A_omeg_word
        .asg            B7,         B_div_inv
        .asg            A17,        A_log
        .asg            A4,         A_a

; ============================================================================
; START:
; ============================ PIPE LOOP PROLOG ==============================
; PROLOG:

        LDBU    .D2T1   *B_ptr_omega--,   A_omega_val                  ;

        NOP             2                                              ;

  [ B_iters]SUB .D2     B_iters,          1,              B_iters      ;
||      GMPY4   .M2     B_denword0,       B_zeroword0,    B_denword0   ;
||      GMPY4   .M1     A_numword1,       A_zeroword1,    A_numword1   ;

  [ B_iters]B   .S2     LOOP1                                          ;
||      PACK2   .L2     B_poly_val,       B_poly_val,     B_poly_word  ;
||      GMPY4   .M1     A_denword1,       A_zeroword1,    A_denword1   ;
||      GMPY4   .M2     B_numword0,       B_zeroword0,    B_numword0   ;
||      LDBU    .D2T2   *B_ptr_poly--,    B_poly_val                   ;

; ============================ PIPE LOOP KERNEL ==============================

*--------------------------------------------------------------------------*
*    for (j = RS_T-1; j >= 0; j--)                                         *
*    {                                                                     *
*         polyval=*ptr_poly --;                                            *
*         polyword=_pack2(polyval,polyval);                                *
*         polyword=_packl4(polyword,polyword);                             *
*                                                                          *
*         omegaval=*ptr_omega --;                                          *
*         omegaword=_pack2(omegaval,omegaval);                             *
*         omegaword=_packl4(omegaword,omegaword);                          *
*                                                                          *
*         denword0=_gmpy4(denword0,zeroword0);                             *
*         denword0=(denword0^polyword);                                    *
*         numword0=_gmpy4(numword0,zeroword0);                             *
*         numword0=(numword0^omegaword);                                   *
*         denword1=_gmpy4(denword1,zeroword1);                             *
*         denword1=(denword1^polyword);                                    *
*         numword1=_gmpy4(numword1,zeroword1);                             *
*         numword1=(numword1^omegaword);                                   *
*    }                                                                     *
*--------------------------------------------------------------------------*


LOOP1:

S_1:                                                                   ; Cycle 1
        PACK2   .L1     A_omega_val,      A_omega_val,    A_omeg_word  ;omeg_word
||      LDBU    .D2T1   *B_ptr_omega--,   A_omega_val                  ;Load

S_2:                                                                   ; Cycle 2
        PACKL4  .L2     B_poly_word,      B_poly_word,    B_poly_word  ;poly_word
||      PACKL4  .L1     A_omeg_word,      A_omeg_word,    A_omeg_word  ;omeg_word

S_3:                                                                   ; Cycle 3
        XOR     .D2     B_denword0,       B_poly_word,    B_denword0   ;dw0^pw
||      XOR     .L1     A_numword1,       A_omeg_word,    A_numword1   ;nw1^om_w

S_4:                                                                   ; Cycle 4
        XOR     .D1X    A_denword1,       B_poly_word,    A_denword1   ;dw1^pw
||      XOR     .L2X    B_numword0,       A_omeg_word,    B_numword0   ;nw0^om_w
||[ B_iters]SUB .D2     B_iters,          1,              B_iters      ;iters--
||      GMPY4   .M2     B_denword0,       B_zeroword0,    B_denword0   ;dw0
||      GMPY4   .M1     A_numword1,       A_zeroword1,    A_numword1   ;nw1

S_5:                                                                   ; Cycle 5
  [ B_iters]B   .S2     LOOP1                                          ;Branch.
||      PACK2   .L2     B_poly_val,       B_poly_val,     B_poly_word  ;poly_word
||      GMPY4   .M1     A_denword1,       A_zeroword1,    A_denword1   ;den_word1
||      GMPY4   .M2     B_numword0,       B_zeroword0,    B_numword0   ;num_word0
||      LDBU    .D2T2   *B_ptr_poly--,    B_poly_val                   ;poly_val

; ============================ PIPE LOOP EPILOG ==============================
; EPILOG:

        PACK2   .L1     A_omega_val,      A_omega_val,   A_omeg_word  ;omeg_word
||[B_iters]LDBU    .D2T1   *B_ptr_omega--,   A_omega_val              ;omega_val

        PACKL4  .L2     B_poly_word,      B_poly_word,   B_poly_word  ;poly_word
||      PACKL4  .L1     A_omeg_word,      A_omeg_word,   A_omeg_word  ;omeg_word
||      LDW     .D1T1   *A_tables[2],     A_log                       ;log

        XOR     .D2     B_denword0,       B_poly_word,   B_denword0   ;denword0
||      XOR     .L1     A_numword1,       A_omeg_word,   A_numword1   ;numword1
||      LDW     .D1T2   *A_tables[1],     B_div_inv                   ;div_inv

        XOR     .D1X    A_denword1,       B_poly_word,   A_denword1   ;denword1
||      XOR     .L2X    B_numword0,       A_omeg_word,   B_numword0   ;numword0
||[ B_iters]SUB .D2     B_iters,          1,             B_iters      ;iters--
||      GMPY4   .M2     B_denword0,       B_zeroword0,   B_denword0   ;denword0
||      GMPY4   .M1     A_numword1,       A_zeroword1,   A_numword1   ;numword1

        PACK2   .L2     B_poly_val,       B_poly_val,    B_poly_word  ;poly_word
||      GMPY4   .M1     A_denword1,       A_zeroword1,   A_denword1   ;denword1
||      GMPY4   .M2     B_numword0,       B_zeroword0,   B_numword0   ;numword0
||      MVK     .S1     0,                A_omega_val                 ;omega_val

        PACK2   .L1     A_omega_val,      A_omega_val,   A_omeg_word  ;omeg_word

        PACKL4  .L2     B_poly_word,      B_poly_word,   B_poly_word  ;poly_word
||      PACKL4  .L1     A_omeg_word,      A_omeg_word,   A_omeg_word  ;omeg_word

        XOR     .D2     B_denword0,       B_poly_word,   B_denword0   ;denword0
||      XOR     .L1     A_numword1,       A_omeg_word,   A_numword1   ;numword1

        XOR     .S1X    A_denword1,       B_poly_word,   A_denword1   ;denword1
||      XOR     .L2X    B_numword0,       A_omeg_word,   B_numword0   ;numword0

; ============================================================================
; END:


*S---------------------------------------------------------------------------*
*S                TI Proprietary Information:  Internal Data                 *
*S---------------------------------------------------------------------------*
*S Swipe6ng datecode: 000104 (04:19:14 PM) [REL]
*S Swipe6ng commandline:
*S swipe6ng -Ma -s -v -l -mv6400 epi.sa
*S Target architecture: "C64xx ISA"
*S---------------------------------------------------------------------------*
*S length = 9
*S Total register usage: 23 (a:12 b:11)
; ====================== SYMBOLIC REGISTER ASSIGNMENTS =======================
        .asg            B17,        B_denword0                  ; 
        .asg            B20,        B_zeroword0                 ; 
        .asg            A9,         A_denword1                  ; 
        .asg            A3,         A_zeroword1                 ; 
        .asg            A10,        A_tables                    ; 
        .asg            B10,        B_T                         ; 
        .asg            A12,        A_N                         ; 
        .asg            B8,         B_fail                      ; 
        .asg            B9,         B_denom_i
        .asg            B0,         B_j                         ; 
        .asg            A19,        A_zero_i
        .asg            A1,         A_numer_i
        .asg            A0,         A_iters                     ; 
        .asg            A18,        A_constant
        .asg            B18,        B_constant
        .asg            A5,         A_thresh
        .asg            B5,         B_4
        .asg            B17,        B_5
        .asg            B7,         B_div_inv
        .asg            A17,        A_log
        .asg            B1,         B_zero                      ; 
        .asg            A7,         A_numword1                  ; 
        .asg            A16,        A_i
        .asg            B16,        B_numword0                  ; 
        .asg            B2,         B_zerval
        .asg            A4,         A_a
        .asg            A6,         A_lam_deg
; ============================================================================
*--------------------------------------------------------------------------*
*    denword0 = _gmpy4(denword0,zeroword0);                                *
*    denword1 = _gmpy4(denword1,zeroword1);                                *
*                                                                          *
*    j = 4;                                                                *
*    denom_i = denword0;                                                   *
*    zero_i  = zeroword0;                                                  *
*    numer_i = numword0;                                                   *
*--------------------------------------------------------------------------*

        GMPY4   .M2     B_denword0,     B_zeroword0,   B_denom_i       ;denom_i 
||      GMPY4   .M1     A_denword1,     A_zeroword1,   A_denword1      ;denword1 
||      MVK     .S1     0FFh,           A_constant                     ;constant
||      MVK     .S2     0FFh,           B_constant                     ;constant
||      MV      .L1X    B_zeroword0,    A_zero_i                       ;zero_i

        MVK     .S2     4,              B_j                            ;j = 4
||      SUB     .L1X    B_T,            1,              A_iters        ;T - 1
||      AND     .L2X    A_zero_i,       B_constant,     B_zerval       ;zerval

        ZERO    .L2     B_zero                                         ;zero
||      ZERO    .L1     A_i                                            ;i
||      MV      .S2     B_j,            B_4                            ;j = 4
||      MV      .D1X    B_numword0,     A_numer_i                      ;numer_i
||      LDBU    .D2T1   *B_div_inv[B_zerval],           A_a            ;a=1/zerval

        ADD     .L2     B_4,           1,              B_5             ; 5
||      SUB     .D1     A_N,           1,              A_thresh        ; N - 1
||      B       .S2     L_3                                            ; Branch.

        ;==== Branch occurs  

*S---------------------------------------------------------------------------*
*S                TI Proprietary Information:  Internal Data                 *
*S---------------------------------------------------------------------------*
*S Swipe6ng datecode: 000104 (04:19:14 PM) [REL]
*S Swipe6ng commandline:
*S swipe6ng -Ma -s -v -2 -mv6400 loop2.cmd
*S Target architecture: "C64xx ISA"
*S---------------------------------------------------------------------------*
*S
*S Resource MII: 7
*S
*S Resource: L1         L1-unit hardware                           Bound: 4
*S Resource: S1         S1-unit hardware                           Bound: 5
*S Resource: D1         D1-unit hardware                           Bound: 6
*S Resource: M1         M1-unit hardware                           Bound: 1
*S Resource: T1         T1-unit LD/ST transport (DAB1)             Bound: 5
*S Resource: 1R         A-side .S1 Long Src / STW .T1 read port    Bound: 1
*S Resource: L2         L2-unit hardware                           Bound: 2
*S Resource: S2         S2-unit hardware                           Bound: 3
*S Resource: D2         D2-unit hardware                           Bound: 5
*S Resource: M2         M2-unit hardware                           Bound: 3
*S Resource: T2         T2-unit LD/ST transport (DAB2)             Bound: 2
*S Resource: 2R         B-side .S2 Long Src / STW .T2 read port    Bound: 2
*S Resource: M1l        M1-unit write port (lo half)               Bound: 1
*S Resource: M2h        M2-unit write port (hi half)               Bound: 3
*S Resource: xoper1     Unshared cross path on 1X                  Bound: 0
*S Resource: xoper2     Unshared cross path on 2X                  Bound: 3
*S Resource: slots      Instruction issue slots: 29                Bound: 4
*S
*S Recurrence MII: 7
*S
*S---------------------------------------------------------------------------*
*S Running adjust phase w/ register allocator
*S---------------------------------------------------------------------------*
*S File   : loop2.cmd
*S ResMII : 7
*S RecMII : 7
*S MII    : 7
*S II     : 7
*S No ILP : 0
*S List   : 25
*S Length : 21
*S Start  : 0
*S Stop   : 20
*S Operations         : 29
*S Avg. para @ II     : 4.142857
*S Calc. Operations   : 29
*S Avg. para calc @ II: 4.142857
*S Live Too Long : 0
*S Unroll Factor : 0
*S---------------------------------------------------------------------------*
*S WARNING: Multiple independent live ranges of 'B_zero' interfere.
*S          Register allocation may be incorrect.
*S Total register usage: 26 (a:17 b:9)
; ====================== SYMBOLIC REGISTER ASSIGNMENTS =======================
        .asg            B9,         B_denom_i                   ; 
        .asg            B18,        B_constant                  ; 
        .asg            A19,        A_zero_i                    ; 
        .asg            A18,        A_constant                  ; 
        .asg            B7,         B_div_inv                   ; 
        .asg            B0,         B_j                         ; 
        .asg            A9,         A_denword1                  ; 
        .asg            A7,         A_numword1                  ; 
        .asg            A3,         A_zeroword1                 ; 
        .asg            B5,         B_4                         ; 
        .asg            A17,        A_log                       ; 
        .asg            A5,         A_thresh                    ; 
        .asg            A16,        A_i                         ; 
        .asg            A6,         A_lam_deg                   ; 
        .asg            B8,         B_fail                      ; 
        .asg            B17,        B_5                         ; 
        .asg            B1,         B_zero                      ; 
        .asg            A0,         A_iters                     ; 
        .asg            A1,         A_numer_i                   ; 
        .asg            B2,         B_denval
        .asg            B2,         B_zerval
        .asg            A21,        A_numval
        .asg            A22,        A_b
        .asg            A4,         A_a
        .asg            A4,         A_err_mag
        .asg            A22,        A_err_pos
        .asg            A2,         A_st
        .asg            A4,         A_st1
        .asg            A20,        A_k
        .asg            A8,         A_byte_i
        .asg            A2,         A_byte_val
        .asg            B29,        B_0
; ============================================================================
; START:
; ============================ PIPE LOOP PROLOG ==============================
; PROLOG:

        SHRU    .S1     A_zero_i,       8,              A_zero_i        ; zero_i
||[ B_j]SUB     .L2     B_j,            1,              B_j             ; j--
||      B       .S2     L_4 + 8                                         ; Branch

  [!B_j]MV      .L1     A_zeroword1,                    A_zero_i        ; zero_i
||      B       .S2     L_5 + 8                                         ; Branch

        NOP             1                                               ; NOP
||      B       .S1     L_6 +  4                                        ; Branch

        SHRU    .S1     A_numer_i,      8,              A_numer_i       ; numer_i 
||      LDBU    .D1T1   *A_log[A_a],                    A_err_pos       ; err_pos
||      AND     .L1     A_numer_i,      A_constant,     A_numval        ; numval
||      AND     .L2X    A_zero_i,       B_constant,     B_zerval        ; zerval
||      B       .S2     L_7                                             ; Branch

  [!B_j]MV      .D1     A_numword1,     A_numer_i                       ; numer_i
||      SHRU    .S2     B_denom_i,      8,              B_denom_i       ; denom_i
||      AND     .L2     B_denom_i,      B_constant,     B_denval        ; denval
||      LDBU    .D2T1   *B_div_inv[B_zerval],           A_a             ;a=1/zerval
||      B       .S1     LOOP2                                           ; Branch.

; ============================ PIPE LOOP KERNEL ==============================
LOOP2:

*--------------------------------------------------------------------------*
*    for(i = 0;i < T; i++)                                                 *
*    {                                                                     *
*        denval = (denom_i & constant);                                    *
*        zerval = (zero_i  & constant);                                    *
*        numval = (numer_i & constant);                                    *
*                                                                          *
*        denom_i = (denom_i >> 8);                                         *
*        zero_i  = (zero_i >> 8);                                          *
*        numer_i = (numer_i >> 8);                                         *
*                                                                          *
*        b = div_inv_table[denval];                                        *
*        j--;                                                              *
*        if(!j) denom_i = denword1;                                        *
*        if(!j) numer_i = numword1;                                        *
*        if(!j) zero_i  = zeroword1;                                       *
*        if(!j) j = 4;                                                     *
*                                                                          *
*        err_mag_i = _gmpy4( numval, b);                                   *
*        err_pos_i = log_table[div_inv_table[zerval]];                     *
*                                                                          *
*        if (!denval)  *fail = 4;                                          *
*        if (err_pos_i > RS_N-1)  *fail = 5;                               *
*                                                                          *
*        k = RS_N-1 - err_pos_i;                                           *
*        if (!*fail) byte_i[k] = ( byte_i[k]^err_mag_i );                  *
*   }                                                                      *
*                                                                          *
*--------------------------------------------------------------------------*

L_1:
        GMPY4   .M1     A_numval,       A_b,            A_err_mag       ; err_mag
||      SHRU    .S1     A_numer_i,      8,              A_numer_i       ; numer_i
||      LDBU    .D1T1   *A_log[A_a],                    A_err_pos       ; err_pos
||      AND     .L1     A_numer_i,      A_constant,     A_numval        ; numval
||      AND     .L2X    A_zero_i,       B_constant,     B_zerval        ; zerval

L_2:
  [ A_iters]B   .S1     LOOP2                                           ; Branch
||      CMPLT   .L1     A_i,            A_lam_deg,      A_st1           ; i<lam_deg
||[ A_st]MPY    .M2     B_5,            1,              B_zero          ; fail=5
||[!B_j]MV      .D1     A_numword1,     A_numer_i                       ; numer_i
||      SHRU    .S2     B_denom_i,      8,              B_denom_i       ; denom_i
||      AND     .L2     B_denom_i,      B_constant,     B_denval        ; denval
||      LDBU    .D2T1   *B_div_inv[B_zerval],           A_a             ; a

L_3:
  [!B_j]MV      .S2X    A_denword1,     B_denom_i                       ;denom_i
||      LDBU    .D2T1   *B_div_inv[B_denval],           A_b             ;b=1/denval

L_4:
  [ A_st]STB    .D2T2   B_5,            *B_fail                         ; *fail =5
||      MPY     .M2X    B_zero,         A_st1,          B_zero          ; zero
||[!B_j]MV      .S2     B_4,            B_j                             ; j = 4

L_5:
        ADD     .D1     A_i,            1,              A_i             ; i++
||      XOR     .L1     A_byte_val,     A_err_mag,      A_byte_val      ;^err_mag
||[!B_denval]STB.D2T2   B_4,            *B_fail                         ;*fail = 0
||[!B_denval]MPY.M2     B_4,            1,              B_zero          ;zero=4
||      SHRU    .S1     A_zero_i,       8,              A_zero_i        ;zero_i
||[ B_j]SUB     .L2     B_j,            1,              B_j             ;j--

L_6:
  [!B_zero]STB  .D1T1   A_byte_val,     *A_byte_i[A_k]                  ;Store
||      SUB     .S1     A_thresh,       A_err_pos,      A_k             ;k   
||[!B_j]MV      .L1     A_zeroword1,    A_zero_i                        ;zero_i

L_7:
  [ A_iters]SUB .S1     A_iters,        1,              A_iters         ;iters--
||      CMPGT   .L1     A_err_pos,      A_thresh,       A_st            ;st
||      LDBU    .D1T1   *A_byte_i[A_k],                 A_byte_val      ;byte_val

; ============================ PIPE LOOP EPILOG ==============================
; EPILOG:

        GMPY4   .M1     A_numval,       A_b,            A_err_mag       ; err_mag
||      B       .S2     B3                                              ; Return

        CMPLT   .L1     A_i,            A_lam_deg,      A_st1           ; st1
||[ A_st]MPY    .M2     B_5,            1,              B_zero          ; zero=5

        NOP             1                                               ; NOP

  [ A_st]STB    .D2T2   B_5,            *B_fail                         ;*fail = 5
||      MPY     .M2X    B_zero,         A_st1,          B_zero          ;

        ADD     .D1     A_i,            1,              A_i             ;
||      XOR     .L1     A_byte_val,     A_err_mag,      A_byte_val      ;^err_mag

  [!B_zero]STB  .D1T1   A_byte_val,     *A_byte_i[A_k]                  ;Store
||         MVC  .S2     B_csr,           CSR                            ; Enable.
; ============================================================================
; END:




* ========================================================================= *
*   End of file:  forney_h.asm                                              *
* ------------------------------------------------------------------------- *
*             Copyright (c) 1999 Texas Instruments, Incorporated.           *
*                            All Rights Reserved.                           *
* ========================================================================= *
