/* ======================================================================== */
/*  NAME                                                                    */
/*                                                                          */
/*      forney -- Forney Algorithm for Reed Solomon Decoder                 */
/*                                                                          */
/*  AUTHOR                                                                  */
/*                                                                          */
/*      J. Sankaran                                                         */
/*                                                                          */
/*  REVISION HISTORY                                                        */
/*                                                                          */
/*      Jun 1998   Natural version of C code.............D.Hoyle            */
/*      01-10-2000 Initial revision of forney . . . . .  J.Sankaran         */
/*      03-08-2000 Removed fail condition 4..............J.Sankaran         */
/*      03-10-2000 Added fail condition 4...............J.Sankaran          */
/*      30-Oct-2000 Make int-c match other flavors . . . .  J. Zbiciak      */
/*                                                                          */
/*  USAGE                                                                   */
/*                                                                          */
/*      This routine has the following C prototype:                         */
/*                                                                          */
/*    void forney_cn(  const  unsigned char restrict   s[],                 */
/*                            unsigned char restrict   lambda[],            */
/*                                     int             lam_deg,             */
/*                            unsigned char restrict   zeros[],             */
/*                            unsigned char restrict   byte_i[],            */
/*                                     int             *fail,               */
/*                      const unsigned char *restrict  tables[],            */
/*                                     int             T,                   */
/*                                     int             RS_N,                */
/*                            unsigned char restrict   scratch[] )          */
/*                                                                          */
/*   T: # of errors that code can correct                                   */
/*   s: array that contains 2T syndromes                                    */
/*   lambda: array of error locator polynomial of degree T + 1              */
/*           where lambda[0] = 1                                            */
/*   lam_deg:Degree of the lambda polynomial.                               */
/*   zeros:  Zeroes of the error locator polynomial found using Chien       */
/*           search.                                                        */
/*   byte_i: received code word of bytes                                    */
/*   fail:   pointer to store status of correction                          */
/*   tables: array of pointers that contain the foll: pointers              */
/*           tables[0] ----> exp_table exponenentials of primitive element  */
/*           tables[1] ----> log_table logarithms of the field elements     */
/*   For eg: exp_table[8] = 2^8 = 29   and inversely log[29] = 8            */
/*   for the case of (204,188,8) code for the default generator polynomial  */
/*   100011101 GF(256)                                                      */
/*   RS_N:   total number of bytes including parity bytes, 204 for the case */
/*           of (204,188,8) code.                                           */
/*   scratch: temporary scratch pad array of size 11 * T to fold temporary  */
/*            results.                                                      */
/*                                                                          */
/*   The forney routine accepts the syndromes array "s" computed using the  */
/*   syndrome routine, "lambda" error locator polynomial computed using the */
/*   "Berley Kamp" or equivalent, "zeros" array that contains the roots of  */
/*   the error locator polynomial, "byte_i" the received code word and      */
/*   corrects the errors in the received codeword array.                    */   
/*                                                                          */
/*  DESCRIPTION                                                             */
/*                                                                          */
/*  Let there be w errors, the zeroes of the lambda polynomial give these   */
/*  locations. These are the roots at which the lambda polynomial evaluates */
/*  to zero. However these are also the same locations where the error poly-*/
/*  nomial does not evaluate to zero. Therefore it can be written as foll:  */
/*  ek . lambdak = 0 for k = 0,1,.....N-1 where N is the same as RS_N       */
/*  Transforming to frequency domain Ek * Lambdak = 0 where * denotes       */
/*  circular convolution. This implies that the dot product of every circul-*/
/*  arly shifted version of Ek with lambda polynomial evaluates to zero.    */
/*  This circular convolution is evaluated over the known coefficients of E */
/*  Since the calculations are performed on a field the linear convolution  */
/*  ends up looking like a circular convolution.  Therefore the linear con- */
/*  voultion equation may be written as: Ek*Lambdak mod(x^(N-1) -1) = 0     */ 
/*  This implies that the product is a multiple of x^n - 1, the multiple    */
/*  being the error evaluator polynomial Omega. Therefore we may write:     */
/*                                                                          */
/*  Ek(x) * Lambdak(x) = Omegak(x).(x^N - 1)                                */
/*                                                                          */
/*  The forney algorithm calculates the first T values of Omega. Knowing the*/
/*  first T values of omega if the number of errors is <= T, coefficients of*/
/*  the error polynomial can be evaluated at these w locations yielding     */
/*                                                                          */
/*  ek = 1/N [(Omega(x)(x^N-1)/Lambda(x)]| for x= alpha^1, alpha^w          */
/*                                                                          */
/*  However these are the same locations at which Lambda(x) is zero, hence  */
/*  the expresssion becomes indeterminate. Therefore by applying L'Hospital */
/*  's rule yields:                                                         */
/*                                                                          */
/*  ek = [alpha^k * Omega(alpha^-k)]/ Lambda'(alpha^-k)                     */
/*                                                                          */
/*  where                                                                   */
/*        ek: error magnitude                                               */
/*        alpha: primitive element                                          */
/*        Omega: Error evaluator polynomial                                 */
/*        Lambda': Derivative of the Lambda polymomial                      */
/*        alpha^-k: zeroes or locations of the errors of the error locator  */
/*                  polynomial                                              */
/*                                                                          */
/*  Error positions are given by log[1/alpha^-k]. Knowing the error locns:  */
/*  and the error magnitudes the errors can be corrected by XOR ing the     */
/*  the error that was added at location k with the error magnitude ek      */
/*                                                                          */
/*                                                                          */
/*  TECHNIQUES                                                              */
/*                                                                          */
/*  Galois field addition is done using XOR's as before, while the multi-   */
/*  plication is done using the GMPY4 instruction. This helps to minimize   */
/*  the difficulty in developing the code.                                  */
/*  The omega polynomial calculation can be completely unrolled. This cal-  */
/*  culation amounts to a circular convolution between the lambda poly-     */
/*  nomial and the syndrome. This can be seen by writing out the omega      */
/*  terms explicitly as follows:                                            */
/*                                                                          */
/*  Omega0 = lam0 * s0                                                      */
/*  Omega1 = lam1 * s0  + lam0 * s1                                         */
/*  Omega2 = lam2 * s0  + lam1 * s1 + lam0 * s2                             */
/*  Omega3 = lam3 * s0  + lam2 * s1 + lam1 * s1 + lam0 * s3                 */
/*  OmegaT = lamT * s0  +....                   + lam0 * sT                 */
/*                                                                          */
/*  Although in theory 2T syndromes are available, only T terms need to be  */
/*  calculated, as the omega polynomial is the error  evaluator polynomial  */
/*  and since there can be utmost T errors T terms of omega suffice.        */
/*                                                                          */
/*  The derivative of the Lambda polynomial is merely the shifted version   */
/*  of the lambda polynomial zeroed out at the even locations. This is true */
/*  because the finite field is of the form GF(2^m). Therefore even terms   */
/*  like 2 evaluate to zero.                                                */
/*                                                                          */
/*  The numerator for the kth error is the omega polynomial evaluated at    */
/*  kth zero of the lambda polynomial and is Omega(alpha^-k). The denomi-   */
/*  nator on the other hand is the derivative of the lambda polynomial      */
/*  evaluated at the kth zero. It is the derivative and not the lambda      */
/*  polynomial itself as by definition the lambda polynomial would          */
/*  evaluate to zero at the kth zero and division by zero is not defined,   */
/*  even in finite fields. The multiplication by alpha^k with the           */
/*  numerator is instead of multiplying the denominator by the zero.        */
/*                                                                          */
/*  ASSUMPTIONS                                                             */
/*                                                                          */
/* a) s: array of syndromes of size 2T                                      */
/* b) lambda: error locator polynomial computed by Berley Kamp or equiv.    */
/* c) lam_deg: degree of error locator polynomial.                          */
/* d) zeros: zeroes of the error locator polynomial found by Chien search.  */
/* e) byte_i: received code word of size N                                  */
/* f) fail: location where fail status is stored                            */
/* g) None of these arrays overlp in memory.                                */
/*                                                                          */
/*                                                                          */
/*  NOTES                                                                   */
/*                                                                          */
/*  None                                                                    */
/*                                                                          */
/*  SOURCE                                                                  */
/*                                                                          */
/* The natural form of C code borrows ideas and concepts derived by the     */
/* following individuals:                                                   */
/*                                                                          */
/* a) Tod Wolf: DSP R&D                                                     */
/* b) David Hoyle: DSP C6000 Software apps group.                           */
/*                                                                          */
/* ------------------------------------------------------------------------ */
/*S                 Adjust the copyright year as necessary.                S*/
/*            Copyright (c) 1999 Texas Instruments, Incorporated.           */
/*                           All Rights Reserved.                           */
/* ======================================================================== */


#pragma CODE_SECTION(forney_c, ".text:intrinsic");
static const char Copyright[] = "Copyright (C) 1999 Texas Instruments "
                                "Incorporated.  All Rights Reserved.";

#include "forney_i.h"

/* ======================================================================== */
/*  forney_c  -- Optimized C with Intrinsics.                               */
/* ======================================================================== */


void forney_c 
(
    const unsigned char  *restrict  s,
    const unsigned char  *restrict  lambda,
    int                             lam_deg,
    const unsigned char  *restrict  zeros,
    unsigned char        *restrict  byte_i,
    int                  *restrict  fail,
    const unsigned char  *restrict
                         *restrict  tables,
    int                  T,
    int                  RS_N,
    unsigned char        *restrict  scratch
)
{
    int i, j, k, fail_code = 0;
    unsigned char *omega = scratch;
    unsigned char *poly  = scratch +  T;
    unsigned char err_pos_i, err_mag_i, b, correct;
    unsigned char constant = 0xFF;

    unsigned int lam_t20, lam_word0;
    unsigned int lam_t31, lam_word1;
    unsigned int lam_t64, lam_word2;
    unsigned int lam_t75, lam_word3;
    unsigned int lam_word4, lam_word00;
    unsigned int lam_word5, lam_word01;
    unsigned int lam_word6;
    unsigned int lam_word7;

    unsigned int *syn_ptr;
    unsigned int sword3210, sword7654;

    unsigned int statew0, statew1;
    unsigned int statew2, statew3;
    unsigned int statew4, statew5;
    unsigned int statew6, statew7;
    unsigned int statew8, statew9;
    unsigned int statew10, statew11;

    unsigned int *omega_ptr = (unsigned int *) (omega);
    unsigned int omega_word0, omega_word1;

    unsigned int *lamptr = (unsigned int *) (lambda);
    unsigned int *poly_ptr = (unsigned int *) (poly);
    unsigned int  poly_word0, poly_word1, poly_word2, poly_word3;

    int  RS_T = T;
  
    const unsigned char *div_inv_table = tables[1];
    const unsigned char *log_table = tables[2];

    double *zerosptr, zero_val;
    unsigned int zeroword0, zeroword1;
    unsigned int denword0,  denword1;
    unsigned int numword0, numword1;

    unsigned char *ptr_omega, *ptr_poly;
    unsigned int   polyval, polyword, omegaval, omegaword;
    unsigned int  denval, numval, zerval;
    unsigned int  denom_i, numer_i, zero_i;

    /*--------------------------------------------------------------*/
    /* Compute omega(x) = S(x)*lambda(x)     Note: the array s[] is */
    /* already in the form of the S(x) syndrome polynomial.         */
    /* omega[0] = lam0 * S0                                         */
    /* omega[1] = lam0 * S1 + lam1 * S0                             */
    /* omega[2] = lam0 * S2 + lam1 * S1 + lam2 * S0                 */
    /* omega[3] = lam0 * S3 + lam1 * S2 + lam2 * S1 + lam3 * S0     */
    /* omega[7] = lam0 * S7 + lam1 * S6 + lam2 * S5 + lam3 * S4     */
    /*           +lam4 * S3 + lam5 * S2 + lam6 * S1                 */
    /*--------------------------------------------------------------*/

    lam_word00 = *lamptr++;  /*lam3210*/
    lam_word01 = *lamptr++;  /*lam7654*/

    lam_t20 = _pack2(lam_word00,  lam_word00);  /* lam20 */
    lam_t31 = _packh2(lam_word00, lam_word00); /* lam31 */
    lam_t64 = _pack2(lam_word01,  lam_word01);  /* lam64 */
    lam_t75 = _packh2(lam_word01, lam_word01); /* lam75 */

    lam_word0 = _packl4(lam_t20, lam_t20);
    lam_word1 = _packh4(lam_t20, lam_t20);
    lam_word2 = _packl4(lam_t31, lam_t31);
    lam_word3 = _packh4(lam_t31, lam_t31);
    lam_word4 = _packl4(lam_t64, lam_t64);
    lam_word5 = _packh4(lam_t64, lam_t64);
    lam_word6 = _packl4(lam_t75, lam_t75);
    lam_word7 = _packh4(lam_t75, lam_t75);

    syn_ptr = (unsigned int *) (s);  
    sword3210   = *syn_ptr++;
    sword7654   = *syn_ptr++;

    statew0     = _gmpy4(lam_word0, sword3210);  /* l0s3 l0s2 l0s1 l0s0  */
    statew1     = _gmpy4(lam_word0, sword7654);  /* l0s7 l0s6 l0s5 l0s4 */

    sword7654   = _shlmb(sword3210, sword7654);  /* sword7654 = sword6543 */
    sword3210   <<= 8;                           /* sword3210 = sword210x */
  
    statew2     = _gmpy4(lam_word1, sword3210);  /* l1s2 l1s1 l1s0  l100 */
    statew3     = _gmpy4(lam_word1, sword7654);  /* l1s6 l1s5 l1s4  l1s3 */

    sword7654   = _shlmb(sword3210, sword7654);  /* sword7654 = sword5432 */
    sword3210   <<= 8;                           /* sword3210 = sword10xx */

    statew4     = _gmpy4(lam_word2, sword3210);  /* l2s1 l2s0 l200  l200 */
    statew5     = _gmpy4(lam_word2, sword7654);  /* l2s5 l2s4 l2s3  l2s2 */

    sword7654   = _shlmb(sword3210, sword7654);  /* sword7654 = sword5432 */
    sword3210   <<= 8;                           /* sword3210 = sword10xx */

    statew6     = _gmpy4(lam_word3, sword3210);  /* l3s0 l3s0 l300  l300 */
    statew7     = _gmpy4(lam_word3, sword7654);  /* l3s4 l3s3 l3s2  l3s1 */

    sword7654   = _shlmb(sword3210, sword7654);  /* sword7654 = sword4321 */
    sword3210   <<= 8;                           /* sword3210 = sword0xxx */

    statew8     = _gmpy4(lam_word4, sword7654);  /* l4s3 l4s2 l4s1  l4s0 */

    sword7654   <<= 8;

    statew9     = _gmpy4(lam_word5, sword7654);  /* l5s2  l5s1 l5s0  l500 */

    sword7654   <<= 8;

    statew10    = _gmpy4(lam_word6, sword7654);  /* l6s1 l6s0 l600  l600 */

    sword7654   <<= 8;

    statew11     = _gmpy4(lam_word7, sword7654);  /* l7s0 l700 l700 l700 */

    omega_word0  = statew0 ^ statew2 ^ statew4 ^ statew6;
    omega_word1  = statew1 ^ statew3 ^ statew5 ^ statew7 ^
                 statew8 ^ statew9 ^ statew10 ^ statew11; 

    *omega_ptr++ = omega_word0;
    *omega_ptr++ = omega_word1;

    /* Now compute the derivative of the lambda poly and store the result */

    poly_word3 = 0;
    poly_word2 = 0;

    poly_word0 = (lam_word00 & 0xFF00FF00);
    poly_word1 = (lam_word01 & 0xFF00FF00);


    *poly_ptr++ = poly_word0;
    *poly_ptr++ = poly_word1;
    *poly_ptr++ = poly_word2;
    *poly_ptr++ = poly_word3;


    /* Now find the error magnitudes and ERROR POSITIONS */

    zerosptr = (double *) (zeros);
    zero_val = *zerosptr;
    zeroword0 = _lo(zero_val);
    zeroword1 = _hi(zero_val);

    denword1 = denword0 = 0;
    numword1 = numword0 = 0;

    ptr_omega = omega + 7; /* pointer to the eight argument */
    ptr_poly  = poly  + 8;  /*pointer to the eight element */

    _nassert(RS_T == 8);
  
    for (j = RS_T-1; j >= 0; j--) 
    {
         polyval=*ptr_poly --;
         polyword=_pack2(polyval,polyval);
         polyword=_packl4(polyword,polyword);

         omegaval=*ptr_omega --;
         omegaword=_pack2(omegaval,omegaval);
         omegaword=_packl4(omegaword,omegaword);

         denword0=_gmpy4(denword0,zeroword0);
         denword0=(denword0^polyword);
         numword0=_gmpy4(numword0,zeroword0);
         numword0=(numword0^omegaword);
         denword1=_gmpy4(denword1,zeroword1);
         denword1=(denword1^polyword);
         numword1=_gmpy4(numword1,zeroword1);
         numword1=(numword1^omegaword);
    }

    denword0 = _gmpy4(denword0,zeroword0);
    denword1 = _gmpy4(denword1,zeroword1);

    j = 4;
    denom_i = denword0;
    zero_i  = zeroword0;
    numer_i = numword0;

    for(i = 0; i < T; i++)
    {
        denval = (denom_i & constant);
        zerval = (zero_i  & constant);
        numval = (numer_i & constant);

        denom_i = (denom_i >> 8);
        zero_i  = (zero_i  >> 8);
        numer_i = (numer_i >> 8);

        b = div_inv_table[denval];
        if (!--j)
        {
          denom_i = denword1;
          numer_i = numword1;
          zero_i  = zeroword1;
          j = 4;
        }

        err_mag_i = _gmpy4(numval, b);
        err_pos_i = log_table[div_inv_table[zerval]];

        if (!denval)             fail_code = 4; 
        if (err_pos_i > RS_N-1)  fail_code = 5; 

        k = RS_N-1 - err_pos_i; /* subtract the error from r(x) */

        correct = byte_i[k] ^ err_mag_i;
        if (!fail_code) 
            byte_i[k] = correct;
   }
   *fail = fail_code;
} 

/* ======================================================================== */
/*  End of file:  forney_i.c                                                */
/* ------------------------------------------------------------------------ */
/*            Copyright (c) 1999 Texas Instruments, Incorporated.           */
/*                           All Rights Reserved.                           */
/* ======================================================================== */
