/* ======================================================================== */
/*  NAME                                                                    */
/*                                                                          */
/*      forney -- Forney Algorithm for Reed Solomon Decoder                 */
/*S                                                                        S*/
/*S AUTHOR                                                                 S*/
/*S     J. Sankaran                                                        S*/
/*S                                                                        S*/
/*S REVISION HISTORY                                                       S*/
/*S     xx-Jun-1998 Natural version of C code  . . . . . .  D.Hoyle        S*/
/*S     10-Jan-2000 Initial revision of forney . . . . . .  J.Sankaran     S*/
/*S     08-Mar-2000 Removed fail condition on denval . . .  J.Sankaran     S*/
/*S     10-Mar-2000 Added fail condition on denval . . . .  J.Sankaran     S*/
/*S     30-Oct-2000 Make opt-c match other flavors . . . .  J. Zbiciak     S*/
/*                                                                          */
/*  USAGE                                                                   */
/*                                                                          */
/*      This routine has the following C prototype:                         */
/*                                                                          */
/*    void forney_cn(  const  unsigned char *restrict  s,                   */
/*                            unsigned char *restrict  lambda,              */
/*                                     int             lam_deg,             */
/*                            unsigned char *restrict  zeros,               */
/*                            unsigned char *restrict  byte_i,              */
/*                                     int  *restrict  *fail,               */
/*                      const unsigned char **restrict tables,              */
/*                                     int             T,                   */
/*                                     int             RS_N,                */
/*                            unsigned char *restrict  scratch)             */
/*                                                                          */
/*   T: # of errors that code can correct                                   */
/*   s: array that contains 2T syndromes                                    */
/*   lambda: array of error locator polynomial of degree T + 1              */
/*           where lambda[0] = 1                                            */
/*   lam_deg:Degree of the lambda polynomial.                               */
/*   zeros:  Zeroes of the error locator polynomial found using Chien       */
/*           search.                                                        */
/*   byte_i: received code word of bytes                                    */
/*   fail:   pointer to store status of correction                          */
/*   tables: array of pointers that contain the foll: pointers              */
/*           tables[0] ----> exp_table exponenentials of primitive element  */
/*           tables[1] ----> log_table logarithms of the field elements     */
/*   For eg: exp_table[8] = 2^8 = 29   and inversely log[29] = 8            */
/*   for the case of (204,188,8) code for the default generator polynomial  */
/*   100011101 GF(256)                                                      */
/*   RS_N:   total number of bytes including parity bytes, 204 for the case */
/*           of (204,188,8) code.                                           */
/*   scratch: temporary scratch pad array of size 11 * T to fold temporary  */
/*            results.                                                      */
/*                                                                          */
/*   The forney routine accepts the syndromes array "s" computed using the  */
/*   syndrome routine, "lambda" error locator polynomial computed using the */
/*   "Berlekamp" or equivalent, "zeros" array that contains the roots of    */
/*   the error locator polynomial, "byte_i" the received code word and      */
/*   corrects the errors in the received codeword array.                    */
/*                                                                          */
/*  DESCRIPTION                                                             */
/*                                                                          */
/*  Let there be w errors, the zeroes of the lambda polynomial give these   */
/*  locations. These are the roots at which the lambda polynomial evaluates */
/*  to zero. However these are also the same locations where the error poly-*/
/*  nomial does not evaluate to zero. Therefore it can be written as foll:  */
/*  ek . lambdak = 0 for k = 0,1,.....N-1 where N is the same as RS_N       */
/*  Transforming to frequency domain Ek * Lambdak = 0 where * denotes       */
/*  circular convolution. This implies that the dot product of every circul-*/
/*  arly shifted version of Ek with lambda polynomial evaluates to zero.    */
/*  This circular convolution is evaluated over the known coefficients of E */
/*  Since the calculations are performed on a field the linear convolution  */
/*  ends up looking like a circular convolution.  Therefore the linear con- */
/*  voultion equation may be written as: Ek*Lambdak mod(x^(N-1) -1) = 0     */
/*  This implies that the product is a multiple of x^n - 1, the multiple    */
/*  being the error evaluator polynomial Omega. Therefore we may write:     */
/*                                                                          */
/*  Ek(x) * Lambdak(x) = Omegak(x).(x^N - 1)                                */
/*                                                                          */
/*  The forney algorithm calculates the first T values of Omega. Knowing the*/
/*  first T values of omega if the number of errors is <= T, coefficients of*/
/*  the error polynomial can be evaluated at these w locations yielding     */
/*                                                                          */
/*  ek = 1/N [(Omega(x)(x^N-1)/Lambda(x)]| for x= alpha^1, alpha^w          */
/*                                                                          */
/*  However these are the same locations at which Lambda(x) is zero, hence  */
/*  the expresssion becomes indeterminate. Therefore by applying L'Hopital's*/
/*  rule yields:                                                            */
/*                                                                          */
/*  ek = [alpha^k * Omega(alpha^-k)]/ Lambda'(alpha^-k)                     */
/*                                                                          */
/*  where                                                                   */
/*        ek: error magnitude                                               */
/*        alpha: primitive element                                          */
/*        Omega: Error evaluator polynomial                                 */
/*        Lambda': Derivative of the Lambda polymomial                      */
/*        alpha^-k: zeroes or locations of the errors of the error locator  */
/*                  polynomial                                              */
/*                                                                          */
/*  Error positions are given by log[1/alpha^-k]. Knowing the error locns:  */
/*  and the error magnitudes the errors can be corrected by XOR ing the     */
/*  the error that was added at location k with the error magnitude ek      */
/*                                                                          */
/*                                                                          */
/*  TECHNIQUES                                                              */
/*                                                                          */
/*  Galois field multiplication is done using logarithms as follows:        */
/*  To do this compuatation addition and multiplication over finite fields  */
/*  is required. Addition over finite fields is merely an XOR operation.    */
/*  The multiplication over finite fields is done using a lookup            */
/*  table for the logarithm of the codewords and an inverse exponentiation  */
/*  table. For eg: if c = a * b where * denotes Galois Field multiplication,*/
/*  then taking logs on both sides we get the following:                    */
/*  log c = log a + log b  = log d                                          */
/*  Taking the exponential amounts to computing the antilog.                */
/*                                                                          */
/*  The Omega polynomial computation is performed by copying the s[2T]      */
/*  array into an array of size 4T starting at location T. The numerator    */
/*  and deno0minator of the error magnitude equation is computed sepe-      */
/*  rately and then evaluated. The location of the error is found by        */
/*  taking the logarithm of the divison of the roots found from Chien       */
/*  search. With the knowledge of the error magnitude and location the      */
/*  receved code word is corrected.                                         */
/*                                                                          */
/*  ASSUMPTIONS                                                             */
/*                                                                          */
/* a) s: array of syndromes of size 2T                                      */
/* b) lambda: error locator polynomial computed by Berley Kamp or equiv.    */
/* c) lam_deg: degree of error locator polynomial.                          */
/* d) zeros: zeroes of the error locator polynomial found by Chien search.  */
/* e) byte_i: received code word of size N                                  */
/* f) fail: location where fail status is stored                            */
/* g) None of these arrays overlp in memory.                                */
/*                                                                          */
/*                                                                          */
/*  NOTES                                                                   */
/*                                                                          */
/*  None                                                                    */
/*                                                                          */
/*  SOURCE                                                                  */
/*                                                                          */
/* The natural form of C code borrows ideas and concepts derived by the     */
/* following individuals:                                                   */
/*                                                                          */
/* a) Tod Wolf: DSP R&D                                                     */
/* b) David Hoyle: DSP C6000 Software apps group.                           */
/*                                                                          */
/* ------------------------------------------------------------------------ */
/*            Copyright (c) 2000 Texas Instruments, Incorporated.           */
/*                           All Rights Reserved.                           */
/* ======================================================================== */

#pragma CODE_SECTION(forney_co,   ".text:optimized");
static const char Copyright[] = "Copyright (c) 2000 Texas Instruments "
                                "Incorporated.  All Rights Reserved.";

#if !defined(_TMS320C6400) && !defined(_TMS320C6500)
#warn Using log-table based GMPY
#include "log_tbl.c"
#define GMPY(a,b) \
              ((a==0)||(b==0)) ? (0):(exp_table2[log_table[a]+log_table[b]])
#else
# define GMPY(a,b) ((unsigned char)_gmpy4((unsigned char)(a), \
                                          (unsigned char)(b)))
#endif


#define MAX_T (8)

#include "forney_o.h"

void forney_co
(
    const unsigned char  *restrict  s,
    const unsigned char  *restrict  lambda,
    int                             lam_deg,
    const unsigned char  *restrict  zeros,
    unsigned char        *restrict  byte_i,
    int                  *restrict  fail,
    const unsigned char  *restrict
                         *restrict  tables,
    int                  RS_T,
    int                  RS_N,
    unsigned char        *restrict  scratch
)
{
    int i,j,k, fail_code = 0;
    unsigned char *restrict omega       = scratch;
    unsigned char *restrict poly        = omega  + RS_T;      
    unsigned char syn[4*MAX_T];
    unsigned char num[  MAX_T];
    unsigned char den[  MAX_T];
    unsigned char err_pos_i, err_mag_i, a, b;

    int  RS_2T = 2*RS_T;

    const unsigned char *restrict div_inv_table = tables[1];
    const unsigned char *restrict log_table     = tables[2];

#ifndef NOASSUME
    _nassert(RS_T    >= 1 && RS_T    <= 8  );
    _nassert(RS_N    >= 1 && RS_N    <= 204);
    _nassert(lam_deg >= 1 && lam_deg <= 8  );
    _nassert((int)lambda  % 4 == 0);
    _nassert((int)zeros   % 8 == 0);
    _nassert((int)s       % 8 == 0);
    _nassert((int)scratch % 8 == 0);
#endif

    for (i = 0           ; i <  RS_T;           i++) syn[i] = 0;
    for (i = RS_T        ; i < (RS_2T + RS_T ); i++) syn[i] = s[i - RS_T];
    for (i = RS_2T + RS_T; i < (RS_2T + RS_2T); i++) syn[i] = 0;

    for (j = RS_T - 1; j >= 0; j--)
    {
        omega[j] = 0;
        for (i = 0; i < RS_T; i++)
            omega[j] ^= GMPY(lambda[i], syn[RS_T + j - i]);
    }

    /* Now compute the derivative of the lambda poly and store the result */
    for (i = 0; i < RS_T; i += 2)
    {
        poly[i    ] = 0;
        poly[i + 1] = lambda[i + 1];
    }
    for (i = RS_T; i< RS_2T; i++) 
        poly[i] = 0;

    /* Now find the error magnitudes and ERROR POSITIONS */
    for (i=0; i< RS_T; i++) 
        den[i] = num[i] = 0;

    for (j = RS_T - 1; j >= 0; j--)
        for (i = 0; i< RS_T; i++)
        {
            den[i] = GMPY(zeros[i], den[i]) ^ poly[1+j];
            num[i] = GMPY(zeros[i], num[i]) ^ omega[j];
        }

    for (i = 0; i < RS_T; i++)
        den[i] = GMPY(den[i], zeros[i]);

    for (i = 0; i < lam_deg; i++)
    {
        b = div_inv_table[den  [i]];
        a = div_inv_table[zeros[i]];

        err_mag_i = GMPY(num[i], b);
        err_pos_i = log_table[a];

        if (!den[i])           fail_code = 4;
        if (err_pos_i >= RS_N) fail_code = 5;

        k = RS_N - 1 - err_pos_i; /* subtract the error from r(x) */

        if (!fail_code)
            byte_i[k] ^= err_mag_i;
    }

    *fail = fail_code;
}

/* ======================================================================== */
/*  End of file:  forney_o.c                                                */
/* ------------------------------------------------------------------------ */
/*            Copyright (c) 2000 Texas Instruments, Incorporated.           */
/*                           All Rights Reserved.                           */
/* ======================================================================== */
